import os
import sys
import logging
import hydra
from pathlib import Path
from omegaconf import DictConfig, OmegaConf
from typing import Optional

sys.path.append("../../")
import vci.train.trainer as train

log = logging.getLogger(__name__)
# os.environ["NCCL_TIMEOUT"] = "36000"

# Custom command line resolver for hydra
import argparse


def main(config_path: Optional[str] = None):
    parser = argparse.ArgumentParser(description="VCI pretraining")
    parser.add_argument("--conf", type=str, help="Path to config YAML file")

    # First parse just the conf argument
    args, override_args = parser.parse_known_args()

    if not args.conf:
        parser.error("--conf argument is required")

    # Initialize hydra with the directory of the config file
    config_file = Path(args.conf)
    config_dir = str(config_file.parent)
    config_name = config_file.name

    # Initialize configuration
    with hydra.initialize_config_module(config_module=None, version_base=None):
        # Load the base configuration
        cfg = OmegaConf.load(args.conf)

        # Process the remaining command line arguments as overrides
        if override_args:
            overrides = OmegaConf.from_dotlist(override_args)
            cfg = OmegaConf.merge(cfg, overrides)

        # Execute the main logic
        run_with_config(cfg)


def run_with_config(cfg: DictConfig):
    if cfg.embeddings.current is None:
        log.error("Gene embeddings are required for training. Please set 'embeddings.current'")
        sys.exit(1)

    if cfg.dataset.current is None:
        log.error("Please set the desired dataset to 'dataset.current'")
        sys.exit(1)

    os.environ["MASTER_PORT"] = str(cfg.experiment.port)
    # WAR: Workaround for sbatch failing when --ntasks-per-node is set.
    # lightning expects this to be set.
    os.environ["SLURM_NTASKS_PER_NODE"] = str(cfg.experiment.num_gpus_per_node)

    log.info(f"*************** Training {cfg.experiment.name} ***************")
    log.info(OmegaConf.to_yaml(cfg))

    train.main(cfg)


if __name__ == "__main__":
    main()



# File: models/decoder_only.py

import torch
from geomloss import SamplesLoss

from .base import PerturbationModel
from .utils import get_activation_class


class DecoderOnlyPerturbationModel(PerturbationModel):
    """
    DecoderOnlyPerturbationModel learns to map the ground truth latent embedding
    (provided in batch["pert_cell_emb"]) to the ground truth HVG space (batch["pert_cell_counts"]).

    Unlike the other perturbation models that compute a control mapping (e.g. via a mapping strategy),
    this model simply feeds the latent representation through a decoder network. The loss is computed
    between the decoder output and the target HVG expression.

    It keeps the overall architectural style (and uses the SamplesLoss loss function from geomloss)
    as in the OldNeuralOT model.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        pert_dim: int,
        n_decoder_layers: int = 2,
        dropout: float = 0.0,
        distributional_loss: str = "energy",
        output_space: str = "gene",
        gene_dim=None,
        **kwargs,
    ):
        super().__init__(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            gene_dim=gene_dim,
            output_dim=output_dim,
            pert_dim=pert_dim,
            output_space=output_space,
            **kwargs,
        )
        self.n_decoder_layers = n_decoder_layers
        self.dropout = dropout
        self.distributional_loss = distributional_loss
        self.cell_sentence_len = kwargs["transformer_backbone_kwargs"]["n_positions"]
        self.activation_class = get_activation_class(kwargs.get("activation", "gelu"))
        self.gene_dim = gene_dim

        # Use the same loss function as OldNeuralOT (e.g. using the MMD loss via geomloss)
        self.loss_fn = SamplesLoss(loss=self.distributional_loss)

    def _build_networks(self):
        pass

    def forward(self, batch: dict) -> torch.Tensor:
        """
        Forward pass: use the ground truth latent embedding (batch["pert_cell_emb"]) as the prediction.
        """
        latent = batch["pert_cell_emb"]
        return latent

    def training_step(self, batch, batch_idx):
        """
        Training step: The decoder output is compared against the target HVG expression.
        We assume that when output_space=="gene", the target is in batch["pert_cell_counts"].
        The predictions and targets are reshaped (using a cell sentence length, if provided)
        before computing the loss.
        """
        pred = self(batch)
        # log a zero tensor
        self.log("train_loss", 0.0)

        if self.gene_decoder is not None and "pert_cell_counts" in batch:
            pert_cell_counts_preds = self.gene_decoder(pred)
            pert_cell_counts_preds = pert_cell_counts_preds.reshape(-1, self.cell_sentence_len, self.gene_dim)
            gene_targets = batch["pert_cell_counts"]
            gene_targets = gene_targets.reshape(-1, self.cell_sentence_len, self.gene_dim)
            decoder_loss = self.loss_fn(pert_cell_counts_preds, gene_targets).mean()
            self.log("decoder_loss", decoder_loss)
        else:
            self.log("decoder_loss", 0.0)
            decoder_loss = None
        return decoder_loss

    def validation_step(self, batch, batch_idx):
        pred = self(batch)
        self.log("val_loss", 0.0)

        return {"loss": None, "predictions": pred}

    def on_validation_batch_end(self, outputs, batch, batch_idx, dataloader_idx=0):
        preds = outputs["predictions"]

        if self.gene_decoder is not None and "pert_cell_counts" in batch:
            pert_cell_counts_preds = self.gene_decoder(preds)
            gene_targets = batch["pert_cell_counts"]
            pert_cell_counts_preds = pert_cell_counts_preds.reshape(-1, self.cell_sentence_len, self.gene_dim)
            gene_targets = gene_targets.reshape(-1, self.cell_sentence_len, self.gene_dim)
            decoder_loss = self.loss_fn(pert_cell_counts_preds, gene_targets).mean()
            self.log("decoder_val_loss", decoder_loss)

    def test_step(self, batch, batch_idx):
        pred = self(batch)

        if self.gene_decoder is not None and "pert_cell_counts" in batch:
            pert_cell_counts_preds = self.gene_decoder(pred)
            gene_targets = batch["pert_cell_counts"]
            gene_targets = gene_targets.reshape(-1, self.cell_sentence_len, self.gene_dim)
            decoder_loss = self.loss_fn(pert_cell_counts_preds, gene_targets).mean()
            self.log("decoder_test_loss", decoder_loss)
        return {"loss": None, "predictions": pred}

    def predict_step(self, batch, batch_idx, padded=True, **kwargs):
        """
        Typically used for final inference. We'll replicate old logic:
         returning 'preds', 'X', 'pert_name', etc.
        """
        latent_output = self.forward(batch)  # shape [B, ...]
        output_dict = {
            "preds": latent_output,
            "pert_cell_emb": batch.get("pert_cell_emb", None),
            "pert_cell_counts": batch.get("pert_cell_counts", None),
            "pert_name": batch.get("pert_name", None),
            "celltype_name": batch.get("cell_type", None),
            "batch": batch.get("batch", None),
            "ctrl_cell_emb": batch.get("ctrl_cell_emb", None),
        }

        pert_cell_counts_preds = self.gene_decoder(latent_output)
        output_dict["pert_cell_counts_preds"] = pert_cell_counts_preds

        return output_dict



import torch
import torch.nn as nn
import torch.nn.functional as F
from geomloss import SamplesLoss


class WassersteinLoss(nn.Module):
    """
    Implements Wasserstein distance loss for distributions represented by logits.
    This implementation supports both 1D and 2D Wasserstein distance calculations.
    """

    def __init__(self, p=1, reduction="mean"):
        """
        Args:
            p (int): Order of Wasserstein distance (1 or 2)
            reduction (str): 'mean', 'sum', or 'none'
        """
        super().__init__()
        self.p = p
        self.reduction = reduction

    def forward(self, p, q):
        """
        Compute Wasserstein distance between predicted and target distributions.

        Args:
            logits (torch.Tensor): Predicted logits of shape (batch_size, num_classes)
            target (torch.Tensor): Target probabilities of shape (batch_size, num_classes)
                                 or class indices of shape (batch_size,)

        Returns:
            torch.Tensor: Computed Wasserstein distance
        """

        q = torch.nan_to_num(q, nan=0.0)
        # Convert logits to probabilities
        pred_probs = F.softmax(p, dim=-1)
        q = F.softmax(q, dim=-1)

        # Compute cumulative distribution functions (CDFs)
        pred_cdf = torch.cumsum(pred_probs, dim=-1)
        target_cdf = torch.cumsum(q, dim=-1)

        max_len = max(pred_cdf.size(1), target_cdf.size(1))
        if pred_cdf.size(1) < max_len:
            pred_cdf = F.pad(pred_cdf, (0, max_len - pred_cdf.size(1)), "constant", 0)
        if target_cdf.size(1) < max_len:
            target_cdf = F.pad(target_cdf, (0, max_len - target_cdf.size(1)), "constant", 0)

        # Compute Wasserstein distance
        wasserstein_dist = torch.abs(pred_cdf - target_cdf).pow(self.p)
        wasserstein_dist = wasserstein_dist.sum(dim=-1)

        # Apply reduction if specified
        if self.reduction == "mean":
            return wasserstein_dist.mean()
        elif self.reduction == "sum":
            return wasserstein_dist.sum()
        return wasserstein_dist


class KLDivergenceLoss(nn.Module):
    def __init__(self, apply_normalization=False, epsilon=1e-10):
        super().__init__()
        self.apply_normalization = apply_normalization
        self.epsilon = epsilon

    def forward(self, p, q):
        q = torch.nan_to_num(q, nan=0.0)
        p = torch.nan_to_num(p, nan=0.0)

        max_len = max(p.size(1), q.size(1))
        if p.size(1) < max_len:
            p = F.pad(p, (0, max_len - p.size(1)), "constant", 0)
        if q.size(1) < max_len:
            q = F.pad(q, (0, max_len - q.size(1)), "constant", 0)

        if self.apply_normalization:
            p = F.softmax(p, dim=-1)
            q = F.softmax(q, dim=-1)

        return torch.sum(p * torch.log(p / q))


class MMDLoss(nn.Module):
    def __init__(self, kernel="energy", blur=0.05, scaling=0.5, downsample=1):
        super().__init__()
        self.mmd_loss = SamplesLoss(loss=kernel, blur=blur, scaling=scaling)
        self.downsample = downsample

    def forward(self, input, target):
        input = input.reshape(-1, self.downsample, input.shape[-1])
        target = target.reshape(-1, self.downsample, target.shape[-1])

        return self.mmd_loss(input, target).mean()


class TabularLoss(nn.Module):
    def __init__(self, shared=128, downsample=1):
        super().__init__()
        self.shared = shared
        self.downsample = downsample

        self.gene_loss = SamplesLoss(loss="energy")
        self.cell_loss = SamplesLoss(loss="energy")

    def forward(self, input, target):
        input = input.reshape(-1, self.downsample, input.shape[-1])
        target = target.reshape(-1, self.downsample, target.shape[-1])
        gene_mmd = self.gene_loss(input, target).mean()

        # cell_mmd should only be on the shared genes, and match scale to mse loss
        cell_inputs = input[:, :, -self.shared :]
        cell_targets = target[:, :, -self.shared :]

        # need to reshape each from (B, self.downsample, F) to (F, self.downsample, B)
        cell_inputs = cell_inputs.transpose(2, 0)
        cell_targets = cell_targets.transpose(2, 0)
        cell_mmd = self.cell_loss(cell_inputs, cell_targets).mean()

        return gene_mmd + cell_mmd



import os
import torch
import lightning as L

from torch import nn
from torch.utils.data import DataLoader
from datetime import timedelta

from lightning.pytorch.callbacks import ModelCheckpoint
from lightning.pytorch.loggers import WandbLogger
from lightning.pytorch.strategies import DDPStrategy
from zclip import ZClipLightningCallback

from ..nn.model import StateEmbeddingModel
from ..data import H5adSentenceDataset, VCIDatasetSentenceCollator
from ..train.callbacks import LogLR, ProfilerCallback, ResumeCallback, EMACallback, PerfProfilerCallback
from ..utils import get_latest_checkpoint, get_embedding_cfg, get_dataset_cfg


def get_embeddings(cfg):
    # Load in ESM2 embeddings and special tokens
    all_pe = torch.load(get_embedding_cfg(cfg).all_embeddings, weights_only=False)
    if isinstance(all_pe, dict):
        all_pe = torch.vstack(list(all_pe.values()))

    all_pe = all_pe.cuda()
    return all_pe


def main(cfg):
    print(f"Starting training with Embedding {cfg.embeddings.current} and dataset {cfg.dataset.current}")
    os.environ["CUDA_LAUNCH_BLOCKING"] = "1"
    os.environ["NCCL_LAUNCH_TIMEOUT"] = str(cfg.experiment.ddp_timeout)
    os.environ["NCCL_ASYNC_ERROR_HANDLING"] = "1"
    TOTAL_N_CELL = cfg.dataset.num_cells
    EPOCH_LENGTH = int(TOTAL_N_CELL // cfg.model.batch_size // 24)
    # ? not sure why this needs to be included but seems empirical?? no clue why this is 6
    warmup_steps = EPOCH_LENGTH * 6

    train_dataset_sentence_collator = VCIDatasetSentenceCollator(cfg, is_train=True)
    # validation should not do augmentations
    val_dataset_sentence_collator = VCIDatasetSentenceCollator(cfg, is_train=False)

    generator = torch.Generator()
    generator.manual_seed(cfg.dataset.seed)

    if get_dataset_cfg(cfg).ds_type == "h5ad":
        DatasetClass = H5adSentenceDataset
    else:
        raise ValueError(f"Unknown dataset type: {get_dataset_cfg(cfg).ds_type}")

    # Training dataloader
    train_dataset = DatasetClass(cfg)
    train_dataloader = DataLoader(
        train_dataset,
        batch_size=cfg.model.batch_size,
        shuffle=True,
        collate_fn=train_dataset_sentence_collator,
        num_workers=cfg.dataset.num_train_workers,
        persistent_workers=True,
        pin_memory=True,
        prefetch_factor=4,
        generator=generator,
    )

    val_dataset = DatasetClass(cfg, test=True)
    val_dataloader = DataLoader(
        val_dataset,
        batch_size=cfg.model.batch_size,
        shuffle=True,
        collate_fn=val_dataset_sentence_collator,
        num_workers=cfg.dataset.num_val_workers,
        persistent_workers=True,
        generator=generator,
    )

    model = StateEmbeddingModel(
        token_dim=get_embedding_cfg(cfg).size,
        d_model=cfg.model.emsize,
        nhead=cfg.model.nhead,
        d_hid=cfg.model.d_hid,
        nlayers=cfg.model.nlayers,
        output_dim=cfg.model.output_dim,
        dropout=cfg.model.dropout,
        warmup_steps=warmup_steps,
        compiled=False,
        max_lr=cfg.optimizer.max_lr,
        emb_size=get_embedding_cfg(cfg).size,
        collater=val_dataset_sentence_collator,
        cfg=cfg,
    )
    model = model.cuda()
    all_pe = get_embeddings(cfg)
    all_pe.requires_grad = False
    model.pe_embedding = nn.Embedding.from_pretrained(all_pe)

    model = model.train()

    run_name, chk = get_latest_checkpoint(cfg)
    checkpoint_callback = ModelCheckpoint(
        every_n_train_steps=cfg.experiment.checkpoint.every_n_train_steps,
        dirpath=os.path.join(cfg.experiment.checkpoint.path, cfg.experiment.name),
        filename=f"{run_name}" + "-{epoch}-{step}",
        save_last=True,
        save_top_k=cfg.experiment.checkpoint.save_top_k,
        monitor=cfg.experiment.checkpoint.monitor,
    )

    if cfg.wandb.enable:
        exp_logger = WandbLogger(project=cfg.wandb.project, name=cfg.experiment.name)
        exp_logger.watch(model, log_freq=1000)
    else:
        exp_logger = None

    callbacks = [checkpoint_callback, LogLR(100), ResumeCallback(cfg), PerfProfilerCallback()]

    if getattr(cfg.model, "ema", False):
        ema_decay = getattr(cfg.model, "ema_decay", 0.999)
        callbacks.append(EMACallback(decay=ema_decay))

    if getattr(cfg.optimizer, "zclip", False):
        zclip_cb = ZClipLightningCallback(
            mode="zscore", alpha=0.97, z_thresh=2.5, clip_option="adaptive_scaling", max_grad_norm=1.0, clip_factor=1.0
        )
        callbacks.append(zclip_cb)

    max_steps = -1
    if cfg.experiment.profile.enable_profiler:
        callbacks.append(ProfilerCallback(cfg=cfg))
        max_steps = cfg.experiment.profile.max_steps

    val_interval = int(cfg.experiment.val_check_interval * cfg.experiment.num_gpus_per_node * cfg.experiment.num_nodes)
    trainer = L.Trainer(
        max_epochs=cfg.experiment.num_epochs,
        max_steps=max_steps,
        callbacks=callbacks,
        devices=cfg.experiment.num_gpus_per_node,
        num_nodes=cfg.experiment.num_nodes,
        # Accumulation
        gradient_clip_val=None if cfg.optimizer.zclip else cfg.optimizer.max_grad_norm,
        accumulate_grad_batches=cfg.optimizer.gradient_accumulation_steps,
        precision="bf16-mixed",
        strategy=DDPStrategy(
            process_group_backend="nccl",
            find_unused_parameters=False,
            timeout=timedelta(seconds=cfg.experiment.get("ddp_timeout", 3600)),
        ),
        val_check_interval=val_interval,
        # Logging
        logger=exp_logger,
        fast_dev_run=False,
        limit_val_batches=cfg.experiment.limit_val_batches,
    )

    if chk:
        print(f"******** Loading chkpoint {run_name} {chk}...")
    else:
        print(f"******** Initialized fresh {run_name}...")

    trainer.fit(model=model, train_dataloaders=train_dataloader, val_dataloaders=val_dataloader, ckpt_path=chk)

    trainer.save_checkpoint(os.path.join(cfg.experiment.checkpoint.path, f"{run_name}_final.pt"))



from typing import Optional

import numpy as np
import torch
import torch.nn as nn
from torch.distributions import Normal
from torch.distributions.kl import kl_divergence as kl
from torch_scatter import scatter_mean
from torchmetrics.functional import pairwise_euclidean_distance, pearson_corrcoef, r2_score
from torchmetrics.functional.clustering import normalized_mutual_info_score

from ._base_modules import CountDecoder, VariationalEncoder
from ._dists import NegativeBinomial, ZeroInflatedNegativeBinomial


def knn_purity(data, labels, n_neighbors=15):
    """Computes KNN Purity for ``data`` given the labels.
    Parameters
    ----------
    data:
        torch tensor of data (n_samples, n_features)
    labels
        torch tensor of labels (n_samples,)
    n_neighbors: int
        Number of nearest neighbors.
    Returns
    -------
    score: float
        KNN purity score. A float between 0 and 1.
    """
    distances = pairwise_euclidean_distance(data)
    # sort each row in distances to get nearest neighbors

    _, indices = torch.topk(distances, k=n_neighbors + 1, dim=1, largest=False, sorted=True)
    indices = indices[:, 1:]  # remove self
    # neighbors_labels = np.vectorize(lambda i: labels[i])(indices)
    neighbors_labels = labels[indices]  # (n_samples, n_neighbors)

    # pre cell purity scores
    scores = ((neighbors_labels - labels.reshape(-1, 1)) == 0).float().mean(axis=1)  # (n_samples,)
    res = scatter_mean(scores, labels).mean()  # per category purity

    return res


class scVIModule(nn.Module):
    """
    scVI module using NegativeBinomial/Zero-Inflated NegativeBinomial Likelihood objectives

    Parameters
    ----------
        n_genes: int
        n_treatments: int
        covars_encoder: dict
            Dictionary of covariates with keys as each covariate name and values as
                number of unique values of the corresponding covariate
        n_latent: int
            Latent Dimension
        loss_ae: str
            Autoencoder loss (either "gauss" or "nb")
        doser_type: str
            # Type of doser network, either `mlp` or `linear`.
        autoencoder_width: int
        autoencoder_depth: int
        use_batch_norm: bool
        use_layer_norm: bool
        variational: bool
    """

    def __init__(
        self,
        n_genes: int,
        n_perts: int,
        n_cell_types: int,
        n_batches: int = 1,
        pert_embeddings: Optional[np.ndarray] = None,
        n_latent: int = 128,
        n_pert_latent: int = 64,
        n_cell_type_latent: int = 32,
        n_batch_latent: int = 16,
        recon_loss: str = "nb",
        n_hidden_encoder: int = 256,
        n_layers_encoder: int = 3,
        n_hidden_decoder: int = 256,
        n_layers_decoder: int = 3,
        use_batch_norm: str = "both",
        use_layer_norm: str = "none",
        dropout_rate_encoder: float = 0.0,
        dropout_rate_decoder: float = 0.0,
        seed: int = 0,
    ):
        super().__init__()

        torch.manual_seed(seed)
        np.random.seed(seed)

        recon_loss = recon_loss.lower()

        assert recon_loss in ["nb", "zinb"]

        self.n_genes = n_genes
        self.n_latent = n_latent
        self.n_cell_type_latent = n_cell_type_latent
        self.n_batch_latent = n_batch_latent
        self.n_pert_latent = n_pert_latent
        self.n_perts = n_perts
        self.recon_loss = recon_loss

        self.encoder = VariationalEncoder(
            n_genes,
            n_latent,
            var_activation=nn.Softplus(),
            n_hidden=n_hidden_encoder,
            n_layers=n_layers_encoder,
            use_batch_norm=use_batch_norm in ["both", "encoder"],
            use_layer_norm=use_layer_norm in ["both", "encoder"],
            dropout_rate=dropout_rate_encoder,
            activation_fn=nn.ReLU,
            return_dist=True,
        )

        n_input_decoder = n_latent + n_cell_type_latent + n_batch_latent + n_pert_latent

        # Decoder components
        if self.recon_loss in ["zinb", "nb"]:
            # setup the parameters of your generative model, as well as your inference model
            self.px_r = torch.nn.Parameter(torch.randn(self.n_genes))

            # decoder goes from n_latent-dimensional space to n_input-d data
            self.decoder = CountDecoder(
                n_input=n_input_decoder,
                n_output=n_genes,
                n_layers=n_layers_decoder,
                n_hidden=n_hidden_decoder,
                use_norm=(
                    "batch"
                    if use_batch_norm in ["both", "decoder"]
                    else "layer"
                    if use_layer_norm in ["both", "decoder"]
                    else "none"
                ),
            )

        elif recon_loss == "gauss":
            self.decoder = VariationalEncoder(
                n_input=n_input_decoder,
                n_output=n_genes,
                n_layers=n_layers_decoder,
                n_hidden=n_hidden_decoder,
                dropout_rate=dropout_rate_decoder,
                use_norm=(
                    "batch"
                    if use_batch_norm in ["both", "decoder"]
                    else "layer"
                    if use_layer_norm in ["both", "decoder"]
                    else "none"
                ),
                var_activation=nn.Softplus(),
            )

        else:
            raise Exception("Invalid Loss function for Autoencoder")

        # Embeddings
        if pert_embeddings is not None:
            self.pert_embeddings = nn.Embedding.from_pretrained(torch.FloatTensor(pert_embeddings), freeze=True)
            self.n_pert_latent = pert_embeddings.shape[1]
        else:
            self.pert_embeddings = nn.Embedding(n_perts, n_pert_latent)

        if n_batches > 1:
            self.batch_embeddings = nn.Embedding(n_batches, n_batch_latent)

        self.cell_type_embeddings = nn.Embedding(n_cell_types, n_cell_type_latent)

        self.metrics = {
            "pearson_r": pearson_corrcoef,
            "r2_score": r2_score,
            "nmi": normalized_mutual_info_score,
        }

    def forward(self, x_basal, perts, cell_types, batch_ids, n_samples: int = 1):
        enc_outputs = self.forward_encoder(
            x=x_basal,
            perts=perts,
            cell_types=cell_types,
            batch_ids=batch_ids,
            n_samples=n_samples,
        )

        dec_outputs = self.forward_decoder(
            z_basal=enc_outputs["z_basal"],
            z=enc_outputs["z"],
            library=enc_outputs["library"],
        )

        return enc_outputs, dec_outputs

    def forward_encoder(
        self,
        x,
        perts,
        cell_types,
        batch_ids: Optional[torch.Tensor] = None,
        n_samples: int = 1,
    ):
        ## TODO: remove unused
        # batch_size = x.shape[0]

        if self.recon_loss in ["nb", "zinb"]:
            # log the input to the variational distribution for numerical stability
            x_ = torch.log(1 + x)
            library = torch.log(x.sum(1)).unsqueeze(1)
        else:
            x_ = x
            library = None, None

        qz, z_basal = self.encoder(x_)

        if n_samples > 1:
            sampled_z = qz.sample((n_samples,))
            z_basal = self.encoder.z_transformation(sampled_z)
            if self.recon_loss in ["nb", "zinb"]:
                library = library.unsqueeze(0).expand((n_samples, library.size(0), library.size(1)))

        z_covs = self.cell_type_embeddings(cell_types.long())
        z_batch = self.batch_embeddings(batch_ids.long())
        z_pert = self.pert_embeddings(perts.long())

        z = torch.cat([z_basal, z_covs, z_batch, z_pert], dim=-1)
        z_corrected = torch.cat([z_basal, z_covs, z_pert], dim=-1)

        return dict(
            z=z,
            z_basal=z_basal,
            z_corrected=z_corrected,
            library=library,
            qz=qz,
        )

    def forward_decoder(
        self,
        z_basal,
        z,
        library,
    ):
        if self.recon_loss == "nb":
            px_scale, _, px_rate, px_dropout = self.decoder("gene", z, library)
            px_r = torch.exp(self.px_r)

            px = NegativeBinomial(mu=px_rate, theta=px_r, scale=px_scale)

        elif self.recon_loss == "zinb":
            px_scale, _, px_rate, px_dropout = self.decoder("gene", z, library)
            px_r = torch.exp(self.px_r)

            px = ZeroInflatedNegativeBinomial(
                mu=px_rate,
                theta=px_r,
                zi_logits=px_dropout,
                scale=px_scale,
            )

        else:
            px_mean, px_var, x_pred = self.decoder(z)

            px = Normal(loc=px_mean, scale=px_var.sqrt())

        pz = Normal(torch.zeros_like(z_basal), torch.ones_like(z_basal))
        return dict(px=px, pz=pz)

    def loss(self, x_pert, encoder_outputs, decoder_outputs):
        """Computes the reconstruction loss (AE) or the ELBO (VAE)"""
        px = decoder_outputs["px"]
        recon_loss = -px.log_prob(x_pert).sum(dim=-1).mean()

        qz = encoder_outputs["qz"]
        pz = decoder_outputs["pz"]

        kl_divergence_z = kl(qz, pz).sum(dim=1)
        kl_loss = kl_divergence_z.mean()

        return recon_loss, kl_loss

    def r2_metric(self, x_pert, x_basal, encoder_outputs, decoder_outputs):
        px = decoder_outputs["px"]
        if self.recon_loss == "gauss":
            x_pred_mean = px.loc

            x_pred_mean = torch.nan_to_num(x_pred_mean, nan=0, posinf=1e3, neginf=-1e3)

            r2_mean = torch.nan_to_num(self.metrics["r2_score"](x_pred_mean.mean(0), x_pert.mean(0)), nan=0.0).item()

            lfc_true = (x_pert - x_basal).mean(0)
            lfc_pred = (x_pred_mean - x_basal).mean(0)

            r2_lfc = torch.nan_to_num(self.metrics["pearson_r"](lfc_pred, lfc_true), nan=0.0).item()

        elif self.recon_loss in ["nb", "zinb"]:
            x_pert = torch.log(1 + x_pert)
            x_pred = px.mu
            x_pred = torch.log(1 + x_pred)
            x_basal = torch.log(1 + x_basal)

            x_pred = torch.nan_to_num(x_pred, nan=0, posinf=1e3, neginf=-1e3)

            r2_mean = torch.nan_to_num(self.metrics["r2_score"](x_pred.mean(0), x_pert.mean(0)), nan=0.0).item()

            lfc_true = (x_pert - x_basal).mean(0)
            lfc_pred = (x_pred - x_basal).mean(0)

            r2_lfc = torch.nan_to_num(self.metrics["pearson_r"](lfc_pred, lfc_true), nan=0.0).item()

        return r2_mean, r2_lfc

    def sample(self, x_basal, perts, cell_types, batch_ids, n_samples=1):
        """Computes knockout gene expression

        Parameters
        ----------
        tensors : dict
            dictionary of input tensors

        """
        _, decoder_outputs = self.forward(
            x_basal=x_basal,
            perts=perts,
            cell_types=cell_types,
            batch_ids=batch_ids,
            n_samples=n_samples,
        )

        px = decoder_outputs["px"]

        if self.recon_loss == "gauss":
            output_key = "loc"
        else:
            output_key = "mu"

        output = getattr(px, output_key)

        return output



from typing import Dict, Tuple

import torch
from torch.optim.lr_scheduler import StepLR

from ..base import PerturbationModel
from ._module import scVIModule


class SCVIPerturbationModel(PerturbationModel):
    """
    Implementation of the scVI model. The outputs are always in
    gene expression space.

    Args:
        input_dim: Dimension of input embeddings (either number of genes or latent dim from obsm key)
        hidden_dim: Dimension of hidden layers
        output_dim: Number of genes to predict
        pert_dim: Dimension of perturbation inputs (usually one-hot size)
        decode_intermediate_dim: Optional intermediate dimension for decoder
        n_encoder_layers: Number of layers in encoder (default: 2)
        n_decoder_layers: Number of layers in encoder (default: 2)
        dropout: Dropout rate (default: 0.1)
        learning_rate: Learning rate for optimizer (default: 1e-3)
        loss_fn: Loss function (default: 'nn.MSELoss()')
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        pert_dim: int,
        n_cell_types: int,
        n_perts: int,
        n_batches: int,
        output_space: str = "gene",
        lr=5e-4,
        wd=1e-6,
        n_steps_kl_warmup: int = None,
        n_epochs_kl_warmup: int = None,
        step_size_lr: int = 45,
        do_clip_grad: bool = False,
        gradient_clip_value: float = 3.0,
        check_val_every_n_epoch: int = 5,
        **kwargs,
    ):
        # Register with parent constructor
        super().__init__(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            output_dim=output_dim,
            pert_dim=pert_dim,
            output_space=output_space,
            **kwargs,
        )

        # Set class specific parameters before registering with parent constructor
        self.n_cell_types = n_cell_types
        self.n_perts = n_perts
        self.n_batches = n_batches

        self.n_layers_encoder = kwargs.get("n_layers_encoder", 2)
        self.n_layers_decoder = kwargs.get("n_layers_decoder", 2)
        self.n_hidden_encoder = kwargs.get("n_hidden_encoder", 256)
        self.n_hidden_decoder = kwargs.get("n_hidden_decoder", 256)
        self.n_latent = kwargs.get("n_latent", 64)
        self.recon_loss = kwargs.get("recon_loss", "nb")

        self.use_batch_norm = kwargs.get("use_batch_norm", "both")
        self.use_layer_norm = kwargs.get("use_layer_norm", "none")

        self.pert_embeddings = None  # will be set in _build_networks

        self.dropout_rate_encoder = kwargs.get("dropout_rate_encoder", 0.0)
        self.dropout_rate_decoder = kwargs.get("dropout_rate_decoder", 0.0)
        self.seed = kwargs.get("seed", 0)

        # training params
        self.lr = lr
        self.wd = wd
        self.n_steps_kl_warmup = n_steps_kl_warmup
        self.n_epochs_kl_warmup = n_epochs_kl_warmup

        self.step_size_lr = step_size_lr
        self.do_clip_grad = do_clip_grad
        self.gradient_clip_value = gradient_clip_value
        self.check_val_every_n_epoch = check_val_every_n_epoch

        self.kwargs = kwargs

        assert self.output_space in ["gene", "all"], "scVI model only supports gene-level or all-level output"

        # Build model components
        self._build_networks()

    def _build_networks(self):
        """
        Build the core components:
        """
        self.module = scVIModule(
            n_genes=self.input_dim,
            n_perts=self.n_perts,
            n_cell_types=self.n_cell_types,
            n_batches=self.n_batches,
            pert_embeddings=self.pert_embeddings,
            n_latent=self.n_latent,
            recon_loss=self.recon_loss,
            n_hidden_encoder=self.n_hidden_encoder,
            n_layers_encoder=self.n_layers_encoder,
            n_hidden_decoder=self.n_hidden_decoder,
            n_layers_decoder=self.n_layers_decoder,
            use_batch_norm=self.use_batch_norm,
            use_layer_norm=self.use_layer_norm,
            dropout_rate_encoder=self.dropout_rate_encoder,
            dropout_rate_decoder=self.dropout_rate_decoder,
            seed=self.seed,
        )

    def encode_perturbation(self, pert: torch.Tensor) -> torch.Tensor:
        """Map perturbation to an effect vector in embedding space."""
        raise NotImplementedError("Perturbation encoding not supported for scVI model")

    def encode_basal_expression(self, expr: torch.Tensor) -> torch.Tensor:
        """Expression is already in embedding space, pass through."""
        raise NotImplementedError("Basal expression encoding not supported for scVI model")

    def perturb(self, pert: torch.Tensor, basal: torch.Tensor) -> torch.Tensor:
        """
        Given a perturbation and basal embeddings, compute the perturbed embedding.
        """
        # Project perturbation and basal cell state to latent space
        raise NotImplementedError("Perturbation not supported for scVI model")

    @property
    def kl_weight(self):
        slope = 1.0
        if self.n_steps_kl_warmup is not None:
            global_step = self.global_step

            if global_step <= self.n_steps_kl_warmup:
                proportion = global_step / self.n_steps_kl_warmup
                return slope * proportion
            else:
                return slope
        elif self.n_epochs_kl_warmup is not None:
            current_epoch = self.current_epoch

            if current_epoch <= self.n_epochs_kl_warmup:
                proportion = current_epoch / self.n_epochs_kl_warmup
                return slope * proportion
            else:
                return slope
        else:
            return slope

    def extract_batch_tensors(
        self, batch: Dict[str, torch.Tensor]
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        x_pert = batch["pert_cell_emb"]
        x_basal = batch["ctrl_cell_emb"]
        pert = batch["pert_emb"]
        cell_type = batch["cell_type_onehot"]
        batch_ids = batch["batch"]

        # if pert is one-hot, convert to index
        if pert.dim() == 2 and pert.size(1) == self.n_perts:
            pert = pert.argmax(1)

        if cell_type.dim() == 2 and cell_type.size(1) == self.n_cell_types:
            cell_type = cell_type.argmax(1)

        if batch_ids.dim() == 2 and batch_ids.size(1) == self.n_batches:
            batch_ids = batch_ids.argmax(1)

        return x_pert, x_basal, pert, cell_type, batch_ids

    def forward(self, batch: Dict[str, torch.Tensor]) -> torch.Tensor:
        """
        Given

        Args:
            batch: Dictionary containing:
                - pert: Perturbation one-hot
                - basal: Control expression embedding
                - cell_type: Cell type one-hot
                - batch: Batch one-hot
        """
        x_pert, x_basal, pert, cell_type, batch_ids = self.extract_batch_tensors(batch)

        encoder_outputs, decoder_outputs = self.module.forward(x_basal, pert, cell_type, batch_ids)

        if self.recon_loss == "gauss":
            output_key = "loc"
        else:
            output_key = "mu"

        output = getattr(decoder_outputs["px"], output_key)

        return output

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        """Training step logic."""
        x_pert, x_basal, pert, cell_type, batch_ids = self.extract_batch_tensors(batch)

        encoder_outputs, decoder_outputs = self.module.forward(x_basal, pert, cell_type, batch_ids)

        recon_loss, kl_loss = self.module.loss(
            x_pert=x_pert,
            encoder_outputs=encoder_outputs,
            decoder_outputs=decoder_outputs,
        )

        loss = recon_loss + self.kl_weight * kl_loss

        r2_mean, r2_lfc = self.module.r2_metric(
            x_pert=x_pert,
            x_basal=x_basal,
            encoder_outputs=encoder_outputs,
            decoder_outputs=decoder_outputs,
        )

        self.log("KL_weight", self.kl_weight, on_epoch=True, on_step=False, prog_bar=True, logger=True)

        self.log(
            "recon_loss",
            recon_loss.item(),
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "r2_mean",
            r2_mean,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "r2_lfc",
            r2_lfc,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        if self.global_step % self.step_size_lr * 1000 == 0:
            sch = self.lr_schedulers()
            sch.step()

        return loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> None:
        """Validation step logic."""
        x_pert, x_basal, pert, cell_type, batch_ids = self.extract_batch_tensors(batch)

        encoder_outputs, decoder_outputs = self.module.forward(x_basal, pert, cell_type, batch_ids)

        recon_loss, kl_loss = self.module.loss(
            x_pert=x_pert,
            encoder_outputs=encoder_outputs,
            decoder_outputs=decoder_outputs,
        )

        r2_mean, r2_lfc = self.module.r2_metric(
            x_pert=x_pert,
            x_basal=x_basal,
            encoder_outputs=encoder_outputs,
            decoder_outputs=decoder_outputs,
        )

        self.log("val_loss", recon_loss + self.kl_weight * kl_loss, prog_bar=True)
        self.log("val_r2_mean", r2_mean, prog_bar=True)
        self.log("val_r2_lfc", r2_lfc, prog_bar=True)

        # is_control = "DMSO_TF" == batch["pert_name"][0] or "non-targeting" == batch["pert_name"][0]
        # if np.random.rand() < 0.1 or is_control:
        #     if self.recon_loss == "gauss":
        #         output_key = "loc"
        #     else:
        #         output_key = "mu"

        #     x_pred = getattr(decoder_outputs["px"], output_key)

        #     self._update_val_cache(batch, x_pred)

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> None:
        x_pert, x_basal, pert, cell_type, batch_ids = self.extract_batch_tensors(batch)

        encoder_outputs, decoder_outputs = self.module.forward(x_basal, pert, cell_type, batch_ids)

        recon_loss, kl_loss = self.module.loss(
            x_pert=x_pert,
            encoder_outputs=encoder_outputs,
            decoder_outputs=decoder_outputs,
        )

        loss = recon_loss + self.kl_weight * kl_loss

        if self.recon_loss == "gauss":
            output_key = "loc"
        else:
            output_key = "mu"

        x_pred = getattr(decoder_outputs["px"], output_key)

        self.log("test_loss", loss, prog_bar=True)
        # self._update_test_cache(batch, x_pred)

        return x_pred

    def predict_step(self, batch, batch_idx, **kwargs):
        """
        Typically used for final inference. We'll replicate old logic:
         returning 'preds', 'X', 'pert_name', etc.
        """
        x_pert, x_basal, pert, cell_type, batch_ids = self.extract_batch_tensors(batch)

        encoder_outputs, decoder_outputs = self.module.forward(x_basal, pert, cell_type, batch_ids)

        if self.recon_loss == "gauss":
            output_key = "loc"
        else:
            output_key = "mu"

        x_pred = getattr(decoder_outputs["px"], output_key)

        return {
            "preds": torch.nan_to_num(torch.log(x_pred + 1), nan=0.0, posinf=1e4, neginf=0.0),
            "pert_cell_counts_preds": torch.nan_to_num(torch.log(x_pred + 1), nan=0.0, posinf=1e4, neginf=0.0),
            "pert_cell_emb": batch.get("pert_cell_emb", None),
            "pert_cell_counts": batch.get("pert_cell_counts", None),
            "pert_emb": batch.get("pert_emb", None),
            "pert_name": batch.get("pert_name", None),
            "cell_type": batch.get("cell_type", None),
            "cell_type_name": batch.get("cell_type_name", None),
            "batch": batch.get("batch", None),
            "batch_name": batch.get("batch_name", None),
            "ctrl_cell_emb": batch.get("ctrl_cell_emb", None),
        }

    def configure_optimizers(self):
        """Set up optimizer."""
        ae_params = (
            list(filter(lambda p: p.requires_grad, self.module.encoder.parameters()))
            + list(filter(lambda p: p.requires_grad, self.module.decoder.parameters()))
            + list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.pert_embeddings.parameters(),
                )
            )
            + list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.cell_type_embeddings.parameters(),
                )
            )
            + list(filter(lambda p: p.requires_grad, self.module.batch_embeddings.parameters()))
        )

        if self.module.recon_loss in ["zinb", "nb"]:
            ae_params += [self.module.px_r]

        optimizer_autoencoder = torch.optim.Adam(ae_params, lr=self.lr, weight_decay=self.wd)

        scheduler_autoencoder = StepLR(optimizer_autoencoder, step_size=self.step_size_lr, gamma=0.9)

        optimizers = [optimizer_autoencoder]
        schedulers = [scheduler_autoencoder]

        if self.step_size_lr is not None:
            return optimizers, schedulers
        else:
            return optimizers



import os
import yaml
import argparse
import logging
import subprocess

from pathlib import Path
from omegaconf import OmegaConf
from hydra import compose, initialize
from jinja2 import Template

log = logging.getLogger(__name__)

logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s %(levelname)s: %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

sbatch_script_template = """#!/bin/bash

#SBATCH --job-name={{ exp_name }}
#SBATCH --nodes={{ num_nodes }}
#SBATCH --gres=gpu:{{ num_gpus_per_node }}
#SBATCH --ntasks-per-node={{ num_gpus_per_node }}
#SBATCH --cpus-per-task=16
#SBATCH --mem=1560G
#SBATCH --time={{ duration }}
#SBATCH --signal=B:SIGINT@300
#SBATCH --output=outputs/{{ exp_name }}/training.log
#SBATCH --open-mode=append
#SBATCH --partition={{ partition }}
{{ sbatch_overrides }}

unset SLURM_TRES_PER_TASK

export MASTER_ADDR=$(scontrol show hostname ${SLURM_JOB_NODELIST} | head -n 1)
export MASTER_PORT='12357'

#export PYTHONFAULTHANDLER=1
#export NCCL_DEBUG=INFO
#export NCCL_DEBUG_SUBSYS=ALL
#export NCCL_VERBOSE_MARK=100
#export TORCH_DISTRIBUTED_DEBUG=DETAIL
#export TORCH_CPP_LOG_LEVEL=INFO

git log --pretty=format:'%h' -n 1

srun \\
    python -m vci.train --conf {{ traing_config_file }}
"""


def parse_vars(extra_vars):
    """
    Parses comma seperated key value pair strings into dict.
    """
    vars_list = []
    if extra_vars:
        for i in extra_vars:
            items = i.split("=")
            key = items[0].strip()
            if len(items) > 1:
                value = "=".join(items[1:])
                vars_list.append((key, value))
    return dict(vars_list)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Create dataset list CSV file")
    parser.add_argument(
        "-c",
        "--config",
        type=str,
        help="Training configuration file.",
    )
    parser.add_argument(
        "-e",
        "--exp_name",
        type=str,
        help="Experiment name. This will be used to name generated artifacts.",
    )
    parser.add_argument(
        "-n",
        "--num_nodes",
        type=int,
        default=1,
        help="Number of nodes to use for this training job.",
    )
    parser.add_argument(
        "-g",
        "--gpus_per_nodes",
        type=int,
        default=4,
        help="Number of GPUs per node",
    )
    parser.add_argument(
        "-r",
        "--reservation",
        dest="reservation",
        type=str,
        default=None,
        help="Slurm reservation to use for this job.",
    )
    parser.add_argument(
        "-p",
        "--partition",
        dest="partition",
        type=str,
        default="gpu_batch,gpu_high_mem,gpu_batch_high_mem,vci_gpu_priority,preemptible",
        help="Slurm partition to use.",
    )
    parser.add_argument(
        "--duration",
        dest="duration",
        type=str,
        default="7-00:00:00",
        help="SLURM job durarion. Pleae refer Slurm documenation for time format",
    )
    parser.add_argument(
        "-f",
        "--force",
        dest="force",
        action="store_true",
        default=False,
        help="Overwrite config and submit the job.",
    )
    parser.add_argument(
        "-d",
        "--dryrun",
        dest="dryrun",
        action="store_true",
        default=False,
        help="Only generate slurm sbatch script",
    )
    parser.add_argument(
        "--set",
        metavar="KEY=VALUE",
        nargs="+",
        default=None,
        help="Values to be overriden for the training.Please refer ./conf/defaults.yaml",
    )

    args = parser.parse_args()

    bind_param = {
        "exp_name": args.exp_name,
        "num_nodes": args.num_nodes,
        "num_gpus_per_node": args.gpus_per_nodes,
        "duration": args.duration,
        "partition": args.partition,
    }

    if args.config:
        bind_param["traing_config_file"] = args.config
    else:
        assert args.exp_name, "Experiment name is required when config is not provided."
        log.info(f"Creating config for {args.exp_name}...")
        trn_conf_dir = Path(f"outputs/{args.exp_name}/conf")
        if not args.force:
            assert not os.path.exists(trn_conf_dir.parent), f"Conf dir {trn_conf_dir.parent.absolute()} already exists."

        overrides = [
            f"experiment.name={args.exp_name}",
            f"experiment.num_nodes={args.num_nodes}",
            f"experiment.num_gpus_per_node={args.gpus_per_nodes}",
        ]

        if args.set:
            log.info(f"Applying overrides: {parse_vars(args.set)}")
            for key, value in parse_vars(args.set).items():
                overrides.append(f"{key}={value}")
            log.info(f"Applying overrides: {overrides}")

        config_dir = Path(os.path.join(os.path.dirname(__file__), "../..", "conf"))
        config_dir = os.path.relpath(config_dir, Path(__file__).parent)
        log.info(config_dir)

        with initialize(version_base=None, config_path=config_dir):
            cfg = compose(
                config_name="defaults.yaml",
                overrides=overrides,
            )
            cfg = OmegaConf.to_container(cfg, resolve=True)

            os.makedirs(trn_conf_dir, exist_ok=True)
            trn_conf_file = Path(f"{trn_conf_dir}/training.yaml")
            with open(trn_conf_file, "w") as file:
                yaml.dump(cfg, file)
            bind_param["traing_config_file"] = trn_conf_file.absolute()

    # SLURM changes
    sbatch_overrides = None
    if args.reservation:
        sbatch_overrides = f"#SBATCH --reservation={args.reservation}\n"

    if sbatch_overrides:
        bind_param["sbatch_overrides"] = sbatch_overrides

    template = Template(sbatch_script_template)
    rendered_script = template.render(bind_param)

    slurm_script = f"outputs/{args.exp_name}/slurm.sh"
    with open(slurm_script, "w") as f:
        f.write(rendered_script)

    if not args.dryrun:
        subprocess.call(["sbatch", slurm_script])



import os
import pickle
import numpy as np
import torch
import anndata as ad
import logging
import yaml

from pathlib import Path
from models import *
from typing import Optional

torch.autograd.set_detect_anomaly(True)
logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)


class InferenceModule:
    def __init__(self, model_folder: str, cell_set_len: Optional[int] = None):
        """
        Initialize the inference module.

        Args:
            model_folder: Path to the folder where the model checkpoint and data_module.pkl are stored.
            model_class: The LightningModule class to use for loading the checkpoint.
        """
        self.model_folder = Path(model_folder)

        # Load the pickled data module to obtain properties.
        dm_path = self.model_folder / "data_module.pkl"
        if not os.path.exists(dm_path):
            raise FileNotFoundError(f"Could not find data_module.pkl in {model_folder}")
        with open(dm_path, "rb") as f:
            self.data_module = pickle.load(f)

        # Obtain embed key and perturbation one-hot map from the loaded data module.
        self.embed_key = self.data_module.embed_key  # e.g. "X_uce"
        self.pert_onehot_map = self.data_module.pert_onehot_map  # dictionary: pert_name -> one-hot tensor
        logger.info(f"Loaded perturbation map with {len(self.pert_onehot_map)} perturbations")

        # Load the model checkpoint.
        self.model: PertSetsPerturbationModel = self._load_model()

        if cell_set_len is None or cell_set_len > self.data_module.cell_sentence_len:
            logger.info(
                f"Requested cell set length {cell_set_len} is None or greater than the maximum cell sentence length "
                f"({self.data_module.cell_sentence_len}). Setting cell_set_len to {self.data_module.cell_sentence_len}"
            )
            self.cell_set_len = self.data_module.cell_sentence_len
        else:
            self.cell_set_len = cell_set_len

    def get_control_pert(self):
        return self.data_module.control_pert

    def available_perturbations(self):
        return list(self.pert_onehot_map.keys())

    def _load_model(self):
        """
        Load the trained model from the checkpoint file.

        Returns:
            The loaded model
        """
        # Find the checkpoint file
        checkpoint_dir = self.model_folder / "checkpoints"
        checkpoint_path = checkpoint_dir / "last.ckpt"

        if not checkpoint_path.exists():
            # the mean baseline models output a "final.ckpt" checkpoint
            checkpoint_backup = checkpoint_dir / "final.ckpt"
            if checkpoint_backup.exists():
                checkpoint_path = checkpoint_backup
            else:
                # Try to find the latest checkpoint based on step number
                checkpoint_files = list(checkpoint_dir.glob("step=*.ckpt"))
                if not checkpoint_files:
                    raise FileNotFoundError(f"No checkpoint files found in {checkpoint_dir}")

                # Sort by step number (assuming filenames contain 'step=X')
                checkpoint_files.sort(key=lambda x: int(str(x).split("step=")[1].split("-")[0]))
                # Get the highest step (last item after sorting)
                checkpoint_path = checkpoint_files[-1]

        logger.info(f"Loading model from checkpoint: {checkpoint_path}")

        # Get model dimensions from the data module
        var_dims = self.data_module.get_var_dims()

        # Determine model type from the config file
        model_config_path = self.model_folder / "config.yaml"
        with open(model_config_path, "r") as f:
            config = yaml.safe_load(f)

        model_class_name = config["model"]["name"]
        ModelClass = self._get_model_class(model_class_name)

        model_kwargs = config["model"]["kwargs"]
        model_init_kwargs = {
            "input_dim": var_dims["input_dim"],
            "hidden_dim": model_kwargs["hidden_dim"],
            "output_dim": var_dims["output_dim"],
            "pert_dim": var_dims["pert_dim"],
            # other model_kwargs keys to pass along:
            **model_kwargs,
        }

        model = ModelClass.load_from_checkpoint(checkpoint_path, strict=False, **model_init_kwargs)
        model.eval()
        return model

    def _get_model_class(self, model_type: str):
        """
        Get the model class based on the model type.

        Args:
            model_type: The type of the model

        Returns:
            The model class
        """
        model_type = model_type.lower()
        model_classes = {
            "simplesum": SimpleSumPerturbationModel,
            "globalsimplesum": GlobalSimpleSumPerturbationModel,
            "celltypemean": CellTypeMeanModel,
            "embedsum": EmbedSumPerturbationModel,
            "neuralot": PertSetsPerturbationModel,  # added as legacy for now for current ckpts. TODO: Remove before release
            "pertsets": PertSetsPerturbationModel,
            "old_neuralot": OldNeuralOTPerturbationModel,
            "decoder_only": DecoderOnlyPerturbationModel,
        }

        if model_type not in model_classes:
            raise ValueError(f"Unknown model type: {model_type}")

        return model_classes[model_type]

    def perturb(self, adata: ad.AnnData, pert_key: str, celltype_key: str) -> ad.AnnData:
        """
        Run inference on an input AnnData and append predictions to .obsm.

        Steps:
        1. Verify that required columns exist.
        2. Remove cells with unknown perturbation labels.
        3. Group cells by (cell_type, perturbation) and split each group into chunks
            (each chunk contains at most cell_set_len cells).
        4. For each chunk, create a batch with:
                - "ctrl_cell_emb": embeddings from adata.obsm[self.embed_key]
                - "pert_emb": repeated one-hot vector for the perturbation.
        5. Run the model (and gene decoder, if available) on each batch.
        6. Restore the original order and add predictions as new keys in .obsm.
        7. Save the modified AnnData.

        Args:
            adata: Input AnnData.
            pert_key: Column name for perturbation labels.
            celltype_key: Column name for cell type labels.

        Returns:
            AnnData with new obsm entries: {embed_key}_pert and X_hvg_pert (if gene decoder exists).
        """
        # Verify required columns.
        if pert_key not in adata.obs.columns:
            raise ValueError(f"Input AnnData must have a column '{pert_key}' in .obs")
        if celltype_key not in adata.obs.columns:
            raise ValueError(f"Input AnnData must have a column '{celltype_key}' in .obs for grouping.")

        # Remove cells with unknown perturbation labels.
        valid_mask = adata.obs[pert_key].isin(self.pert_onehot_map)
        if valid_mask.sum() < adata.n_obs:
            logger.info(f"Removing {adata.n_obs - valid_mask.sum()} cells with unknown perturbation labels.")
            adata = adata[valid_mask].copy()

        # Get cell embeddings.
        X_embed = adata.obsm[self.embed_key]
        if not torch.is_tensor(X_embed):
            X_embed = torch.tensor(X_embed, dtype=torch.float32)
        device = next(self.model.parameters()).device
        X_embed = X_embed.to(device)

        # Extract perturbation and cell type labels; store original indices.
        pert_labels = adata.obs[pert_key].values
        cell_types = adata.obs[celltype_key].values
        original_indices = np.arange(adata.n_obs)

        # Group cells by (cell_type, perturbation).
        groups = {}
        for i, (ct, p) in enumerate(zip(cell_types, pert_labels)):
            key = (ct, p)
            groups.setdefault(key, []).append(i)

        latent_predictions = {}
        gene_predictions = {}

        from tqdm import tqdm

        # Loop over groups with a progress bar.
        for (ct, p), idx_list in tqdm(list(groups.items()), total=len(groups), desc="Processing groups"):
            onehot = self.pert_onehot_map[p].to(device)
            n = len(idx_list)
            # Process each group in chunks (cell sentences).
            for start in tqdm(range(0, n, self.cell_set_len), desc=f"Group {ct}-{p}", leave=False):
                chunk_indices = idx_list[start : start + self.cell_set_len]
                basal = X_embed[chunk_indices, :]  # (chunk_size, embed_dim)
                pert_tensor = onehot.unsqueeze(0).repeat(len(chunk_indices), 1)
                pert_names = [p] * len(chunk_indices)
                batch = {"ctrl_cell_emb": basal, "pert_emb": pert_tensor, "pert_name": pert_names}
                with torch.no_grad():
                    # if self.model is class PertSetsPerturbationModel, need to set padded=False
                    if isinstance(self.model, PertSetsPerturbationModel):
                        batch_pred = self.model.forward(batch, padded=False)
                    else:
                        batch_pred = self.model.forward(batch)
                    latent_out = batch_pred
                    gene_out = (
                        self.model.gene_decoder(latent_out)
                        if (hasattr(self.model, "gene_decoder") and self.model.gene_decoder is not None)
                        else None
                    )
                latent_np = latent_out.detach().cpu().numpy()
                if gene_out is not None:
                    gene_np = gene_out.detach().cpu().numpy()
                for i_local, orig_idx in enumerate(chunk_indices):
                    latent_predictions[orig_idx] = latent_np[i_local]
                    if gene_out is not None:
                        gene_predictions[orig_idx] = gene_np[i_local]

        # Reconstruct predictions in the original order.
        valid_indices = sorted(latent_predictions.keys())
        latent_array = np.array([latent_predictions[i] for i in valid_indices])
        gene_array = np.array([gene_predictions[i] for i in valid_indices]) if gene_out is not None else None

        adata_out = adata[valid_indices].copy()
        adata_out.obsm[self.embed_key + "_pert"] = latent_array
        if gene_array is not None:
            adata_out.obsm["X_hvg_pert"] = gene_array
        return adata_out



# File: vci/flash_transformer.py
"""
This module implements a Transformer encoder layer.
"""

import torch
import torch.nn as nn
import torch.nn.functional as F


class FlashTransformerEncoderLayer(nn.Module):
    def __init__(self, d_model, nhead, dim_feedforward, dropout=0.1):
        """
        Initializes the encoder layer.
        Args:
            d_model (int): model dimension.
            nhead (int): number of attention heads.
            dim_feedforward (int): dimension of the feed-forward network.
            dropout (float): dropout probability.
        """
        super().__init__()
        torch.backends.cuda.enable_flash_sdp(True)

        self.d_model = d_model
        self.nhead = nhead
        self.dropout = dropout

        # Linear projections for Q, K, V in one matrix
        self.qkv_proj = nn.Linear(d_model, d_model * 3)
        self.out_proj = nn.Linear(d_model, d_model)

        self.norm1 = nn.LayerNorm(d_model)
        self.norm2 = nn.LayerNorm(d_model)
        self.dropout_layer = nn.Dropout(dropout)

        # Feed-forward network
        self.linear1 = nn.Linear(d_model, dim_feedforward)
        self.linear2 = nn.Linear(dim_feedforward, d_model)

    def forward(self, src, src_mask=None, src_key_padding_mask=None):
        """
        Args:
            src: Tensor of shape (batch_size, seq_len, d_model)
            src_mask: (optional) attention mask.
            src_key_padding_mask: (optional) padding mask.
        Returns:
            Tensor of shape (batch_size, seq_len, d_model)
        """
        # For this simple implementation, we'll use either one of the masks.
        # You can combine them as needed.
        mask = src_key_padding_mask if src_key_padding_mask is not None else src_mask

        # ----- Self-Attention Block -----
        residual = src

        # Compute Q, K, V projections in one go.
        qkv = self.qkv_proj(src)  # shape: (B, T, 3*d_model)
        q, k, v = torch.chunk(qkv, 3, dim=-1)  # each: (B, T, d_model)

        # Reshape for multi-head attention.
        head_dim = self.d_model // self.nhead
        q = q.view(src.size(0), src.size(1), self.nhead, head_dim).transpose(1, 2)  # (B, nhead, T, head_dim)
        k = k.view(src.size(0), src.size(1), self.nhead, head_dim).transpose(1, 2)
        v = v.view(src.size(0), src.size(1), self.nhead, head_dim).transpose(1, 2)

        # Use PyTorch’s built-in scaled_dot_product_attention.
        attn_output = F.scaled_dot_product_attention(q, k, v, dropout_p=self.dropout, is_causal=False)
        # Merge heads.
        attn_output = attn_output.transpose(1, 2).contiguous().view(src.size(0), src.size(1), self.d_model)
        attn_output = self.out_proj(attn_output)
        src = self.norm1(residual + self.dropout_layer(attn_output))

        # ----- Feed-Forward Block -----
        residual2 = src
        ff_output = self.linear2(self.dropout_layer(F.gelu(self.linear1(src))))
        src = self.norm2(residual2 + self.dropout_layer(ff_output))
        return src


class FlashTransformerEncoder(nn.Module):
    def __init__(self, layers):
        """
        A simple encoder that applies a stack of FlashTransformerEncoderLayer instances.
        Args:
            layers (list[nn.Module]): list of FlashTransformerEncoderLayer instances.
        """
        super().__init__()
        self.layers = nn.ModuleList(layers)

    def forward(self, src, src_mask=None, src_key_padding_mask=None):
        """
        Applies each encoder layer in sequence.
        Args:
            src: Tensor of shape (B, T, d_model)
            src_mask: (optional) attention mask.
            src_key_padding_mask: (optional) padding mask.
        Returns:
            Tensor of shape (B, T, d_model)
        """
        # Use src_key_padding_mask if provided; otherwise use src_mask.
        mask = src_key_padding_mask if src_key_padding_mask is not None else src_mask
        output = src
        for layer in self.layers:
            output = layer(output, src_mask=mask, src_key_padding_mask=mask)
        return output



from ._emb import add_arguments_emb, run_emb_fit, run_emb_transform
from ._tx import add_arguments_tx, run_tx_infer, run_tx_predict, run_tx_train

__all__ = [
    "add_arguments_emb",
    "add_arguments_tx",
    "run_tx_train",
    "run_tx_predict",
    "run_tx_infer",
    "run_emb_fit",
    "run_emb_transform",
]



import logging
from collections import defaultdict
from typing import Dict, Tuple

import torch
import torch.nn as nn

from .base import PerturbationModel

logger = logging.getLogger(__name__)


class GlobalSimpleSumPerturbationModel(PerturbationModel):
    """
    A map-independent baseline that computes a cell-type-specific
    control mean and a cell-type-specific offset for each (cell type, perturbation).

    Implementation details:
      - We do a single pass over the training dataloader (in on_fit_start) to
        accumulate sums & counts:
        * celltype_ctrl_sum[c_name] and celltype_ctrl_count[c_name]
        * celltype_pert_sum[c_name][p_name] and celltype_pert_count[c_name][p_name]
      - Then we compute:
          ctrl_mean[c_name] = celltype_ctrl_sum[c_name]/celltype_ctrl_count[c_name]
          offset[(c_name, p_name)] = (pert_sum / pert_count) - ctrl_mean[c_name]
      - At inference, for each sample with cell type c and perturbation p,
        we predict: ctrl_mean[c] + offset[(c, p)].
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        pert_dim: int,
        dropout: float = 0.0,
        lr: float = 1e-3,
        loss_fn=nn.MSELoss(),
        embed_key: str = None,
        output_space: str = "gene",
        gene_names=None,
        **kwargs,
    ):
        """
        Args:
            input_dim: Size of input features (genes or embedding).
            hidden_dim: Not used here, but required by base-class signature.
            output_dim: Dimension of the output (often #genes).
            pert_dim:  Dimension of perturbation embeddings (not used here).
            n_decoder_layers: (Unused) included for config compatibility.
            dropout: (Unused) included for config compatibility.
            lr: Learning rate if there's anything to optimize. We only keep a dummy param.
            loss_fn: The chosen PyTorch loss function for training (default MSE).
            embed_key: Possibly an embedding key from the datamodule (unused).
            output_space: 'gene' or 'latent' (we handle either; just read from batch["pert_cell_counts"] or batch["pert_cell_emb"]).
            decoder: Possibly a separate decoder if output_space='latent' for final evaluation.
            gene_names: Names of genes if needed for logging.
            kwargs: Catch-all for extra arguments from config.
        """
        super().__init__(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            output_dim=output_dim,
            pert_dim=pert_dim,
            dropout=dropout,
            lr=lr,
            loss_fn=loss_fn,
            embed_key=embed_key,
            output_space=output_space,
            gene_names=gene_names,
            **kwargs,
        )

        # We'll store a mean control vector per cell type:
        self.ctrl_mean: Dict[str, torch.Tensor] = {}
        # We'll store an offset for each (cell_type, pert_name):
        self.offsets: Dict[Tuple[str, str], torch.Tensor] = {}

        # A dummy parameter so that Lightning sees something to "optimize"
        self.dummy_param = nn.Parameter(torch.zeros(1, requires_grad=True))
        self.pert_mean_offsets = {}

    def on_fit_start(self):
        """Called by Lightning before training."""
        super().on_fit_start()

        # 1. First compute control means per cell type
        celltype_ctrl_means = {}  # Dict[str, torch.Tensor]

        # 2. And perturbation means per cell type
        celltype_pert_means = defaultdict(dict)  # Dict[str, Dict[str, torch.Tensor]]

        train_loader = self.trainer.datamodule.train_dataloader()
        if train_loader is None:
            logger.warning("No train dataloader found. Cannot compute offsets.")
            return

        # First pass: gather sums per cell type
        celltype_sums = defaultdict(lambda: defaultdict(lambda: {"sum": torch.zeros(self.output_dim), "count": 0}))

        with torch.no_grad():
            for batch in train_loader:
                if (
                    self.embed_key
                    and self.embed_key != "X_hvg"
                    and self.output_space == "gene"
                    or self.embed_key
                    and self.output_space == "all"
                ):
                    X_vals = batch["pert_cell_counts"]
                else:
                    X_vals = batch["pert_cell_emb"]

                X_cpu = X_vals.float().cpu()
                pert_names = batch["pert_name"]
                cell_types = batch["cell_type"]

                for i in range(len(X_cpu)):
                    p_name = str(pert_names[i])
                    ct_name = str(cell_types[i])
                    x_val = X_cpu[i]

                    celltype_sums[ct_name][p_name]["sum"] += x_val
                    celltype_sums[ct_name][p_name]["count"] += 1

            # Now compute means per cell type
            all_ctrl_means = []  # For computing global basal
            for ct_name, pert_dict in celltype_sums.items():
                # Get control mean for this cell type
                ctrl_stats = pert_dict.get(self.control_pert)
                if ctrl_stats is None or ctrl_stats["count"] == 0:
                    logger.warning(f"No control cells found for cell type {ct_name}")
                    continue

                ct_ctrl_mean = ctrl_stats["sum"] / ctrl_stats["count"]
                celltype_ctrl_means[ct_name] = ct_ctrl_mean
                all_ctrl_means.append(ct_ctrl_mean)

                # Compute perturbation means and deltas for this cell type
                for p_name, stats in pert_dict.items():
                    if p_name == self.control_pert:
                        continue

                    if stats["count"] == 0:
                        continue

                    ct_pert_mean = stats["sum"] / stats["count"]
                    ct_delta = ct_pert_mean - ct_ctrl_mean
                    celltype_pert_means[ct_name][p_name] = ct_delta

            # Now average deltas across cell types
            self.pert_mean_offsets = {}
            for p_name in set().union(*[d.keys() for d in celltype_pert_means.values()]):
                # Gather deltas for this perturbation across cell types
                pert_deltas = []
                for ct_name, pert_dict in celltype_pert_means.items():
                    if p_name in pert_dict:
                        pert_deltas.append(pert_dict[p_name])

                if not pert_deltas:
                    continue

                # Average the deltas
                self.pert_mean_offsets[p_name] = torch.stack(pert_deltas).mean(0)

            # Add zero offset for control
            self.pert_mean_offsets[self.control_pert] = torch.zeros(self.output_dim)

            # Compute global basal as mean of cell-type means
            if not all_ctrl_means:
                logger.warning("No control cells found in any cell type. Using zero vector as basal.")
                self.global_basal = torch.zeros(self.output_dim)
            else:
                self.global_basal = torch.stack(all_ctrl_means).mean(0)

            logger.info(f"GlobalSimpleSum: computed offsets for {len(self.pert_mean_offsets)} perturbations")

    def forward(self, batch: dict) -> torch.Tensor:
        """
        Now we ignore the cell type entirely.

        For each sample in the batch, we do:
            prediction = global_basal + offset[p_name].
        """
        B = len(batch["pert_name"])
        device = self.dummy_param.device
        pred_out = torch.zeros((B, self.output_dim), device=device)

        for i in range(B):
            p_name = str(batch["pert_name"][i])
            offset_vec = self.pert_mean_offsets.get(p_name, None)
            if offset_vec is None:
                offset_vec = torch.zeros(self.output_dim, device=device)

            pred_out[i] = batch["ctrl_cell_emb"][i] + offset_vec.to(device)

        return pred_out

    def configure_optimizers(self):
        """
        Return an optimizer for our dummy_param if desired, or None if we prefer no updates.
        """
        if len(list(self.parameters())) > 0:
            return torch.optim.Adam(self.parameters(), lr=self.lr)
        return None

    def training_step(self, batch, batch_idx):
        """
        We'll compute MSE vs. the ground truth. (Though no real learnable offsets.)
        """
        pred = self(batch)
        if (self.embed_key and self.embed_key != "X_hvg" and self.output_space == "gene") or (
            self.embed_key and self.output_space == "all"
        ):
            target = batch["pert_cell_counts"]
        else:
            target = batch["pert_cell_emb"]
        loss = self.loss_fn(pred, target)
        self.log("train_loss", loss, prog_bar=True)
        return None

    def on_save_checkpoint(self, checkpoint):
        """
        Save the global_basal vector and perturbation offsets to the checkpoint.

        Args:
            checkpoint (dict): Checkpoint dictionary to be saved.
        """
        super().on_save_checkpoint(checkpoint)

        # Convert tensors to CPU and NumPy arrays for serialization
        checkpoint["global_basal"] = self.global_basal.cpu().numpy()
        checkpoint["pert_mean_offsets"] = {
            p_name: offset.cpu().numpy() for p_name, offset in self.pert_mean_offsets.items()
        }

        logger.info("GlobalSimpleSum: Saved global_basal and pert_mean_offsets to checkpoint.")

    def on_load_checkpoint(self, checkpoint):
        """
        Load the global_basal vector and perturbation offsets from the checkpoint.

        Args:
            checkpoint (dict): Checkpoint dictionary from which to load.
        """
        super().on_load_checkpoint(checkpoint)

        # Load global_basal
        if "global_basal" in checkpoint:
            self.global_basal = torch.tensor(checkpoint["global_basal"], dtype=torch.float32)
            logger.info("GlobalSimpleSum: Loaded global_basal from checkpoint.")
        else:
            logger.warning("GlobalSimpleSum: No global_basal found in checkpoint. Using zero vector.")
            self.global_basal = torch.zeros(self.output_dim)

        # Load perturbation offsets
        if "pert_mean_offsets" in checkpoint:
            loaded_offsets = {}
            for p_name, offset_np in checkpoint["pert_mean_offsets"].items():
                loaded_offsets[p_name] = torch.tensor(offset_np, dtype=torch.float32)
            self.pert_mean_offsets = loaded_offsets
            logger.info(
                f"GlobalSimpleSum: Loaded offsets for {len(self.pert_mean_offsets)} perturbations from checkpoint."
            )
        else:
            logger.warning("GlobalSimpleSum: No pert_mean_offsets found in checkpoint. All offsets set to zero.")
            self.pert_mean_offsets = {}

    def encode_perturbation(self, pert: torch.Tensor) -> torch.Tensor:
        """Not really used here, but required by abstract base. We do no param-based encoding."""
        return pert

    def encode_basal_expression(self, expr: torch.Tensor) -> torch.Tensor:
        """No param-based encoding of basal. Just identity."""
        return expr

    def perturb(self, pert: torch.Tensor, basal: torch.Tensor) -> torch.Tensor:
        """
        Not used in the normal forward pass, because we look up offset by 'pert_name' strings.
        But we must define it to meet the abstract contract.
        """
        return basal

    def _build_networks(self):
        """
        We don't need any networks to be built.
        """
        pass



import argparse


def add_arguments_infer(parser: argparse.ArgumentParser):
    parser.add_argument(
        "--checkpoint",
        type=str,
        required=False,
        help="Path to model checkpoint (.ckpt). If not provided, will use model_dir/checkpoints/final.ckpt",
    )
    parser.add_argument("--adata", type=str, required=True, help="Path to input AnnData file (.h5ad)")
    parser.add_argument("--embed_key", type=str, default="X_hvg", help="Key in adata.obsm for input features")
    parser.add_argument(
        "--pert_col", type=str, default="drugname_drugconc", help="Column in adata.obs for perturbation labels"
    )
    parser.add_argument("--output", type=str, default=None, help="Path to output AnnData file (.h5ad)")
    parser.add_argument(
        "--model_dir",
        type=str,
        required=True,
        help="Path to the model_dir containing the config.yaml file and the pert_onehot_map.pt file that was saved during training.",
    )
    parser.add_argument(
        "--celltype_col", type=str, default=None, help="Column in adata.obs for cell type labels (optional)"
    )
    parser.add_argument(
        "--celltypes", type=str, default=None, help="Comma-separated list of cell types to include (optional)"
    )
    parser.add_argument("--batch_size", type=int, default=1000, help="Batch size for inference (default: 1000)")


def run_tx_infer(args):
    import logging
    import os
    import pickle

    import numpy as np
    import scanpy as sc
    import torch
    import yaml
    from tqdm import tqdm

    from ...sets.models.pert_sets import PertSetsPerturbationModel

    logging.basicConfig(level=logging.INFO)
    logger = logging.getLogger(__name__)

    def load_config(cfg_path: str) -> dict:
        """Load config from the YAML file that was dumped during training."""
        if not os.path.exists(cfg_path):
            raise FileNotFoundError(f"Could not find config file: {cfg_path}")
        with open(cfg_path, "r") as f:
            cfg = yaml.safe_load(f)
        return cfg

    # Load the config
    config_path = os.path.join(args.model_dir, "config.yaml")
    cfg = load_config(config_path)
    logger.info(f"Loaded config from {config_path}")

    # Determine checkpoint path
    if args.checkpoint is not None:
        checkpoint_path = args.checkpoint
    else:
        checkpoint_path = os.path.join(args.model_dir, "checkpoints", "final.ckpt")
        logger.info(f"No checkpoint provided, using default: {checkpoint_path}")

    # Get perturbation dimensions and mapping from data module
    var_dims_path = os.path.join(args.model_dir, "var_dims.pkl")
    with open(var_dims_path, "rb") as f:
        var_dims = pickle.load(f)
    pert_dim = var_dims["pert_dim"]

    # Load model
    logger.info(f"Loading model from checkpoint: {checkpoint_path}")
    model = PertSetsPerturbationModel.load_from_checkpoint(checkpoint_path)
    model.eval()
    cell_sentence_len = model.cell_sentence_len
    device = next(model.parameters()).device

    # Load AnnData
    logger.info(f"Loading AnnData from: {args.adata}")
    adata = sc.read_h5ad(args.adata)

    # Optionally filter by cell type
    if args.celltype_col is not None and args.celltypes is not None:
        celltypes = [ct.strip() for ct in args.celltypes.split(",")]
        if args.celltype_col not in adata.obs:
            raise ValueError(f"Column '{args.celltype_col}' not found in adata.obs.")
        initial_n = adata.n_obs
        adata = adata[adata.obs[args.celltype_col].isin(celltypes)].copy()
        logger.info(f"Filtered AnnData to {adata.n_obs} cells of types {celltypes} (from {initial_n} cells)")
    elif args.celltype_col is not None:
        if args.celltype_col not in adata.obs:
            raise ValueError(f"Column '{args.celltype_col}' not found in adata.obs.")
        logger.info(f"No cell type filtering applied, but cell type column '{args.celltype_col}' is available.")

    # Get input features
    if args.embed_key in adata.obsm:
        X = adata.obsm[args.embed_key]
        logger.info(f"Using adata.obsm['{args.embed_key}'] as input features: shape {X.shape}")
    else:
        X = adata.X
        logger.info(f"Using adata.X as input features: shape {X.shape}")

    # Prepare perturbation tensor using the data module's mapping
    pert_names = adata.obs[args.pert_col].values
    pert_tensor = torch.zeros((len(pert_names), pert_dim), device="cpu")  # Keep on CPU initially
    logger.info(f"Perturbation tensor shape: {pert_tensor.shape}")

    # Load perturbation mapping from torch file
    pert_onehot_map_path = os.path.join(args.model_dir, "pert_onehot_map.pt")
    pert_onehot_map = torch.load(pert_onehot_map_path, weights_only=False)

    logger.info(f"Data module has {len(pert_onehot_map)} perturbations in mapping")
    logger.info(f"First 10 perturbations in data module: {list(pert_onehot_map.keys())[:10]}")

    unique_pert_names = sorted(set(pert_names))
    logger.info(f"AnnData has {len(unique_pert_names)} unique perturbations")
    logger.info(f"First 10 perturbations in AnnData: {unique_pert_names[:10]}")

    # Check overlap
    overlap = set(unique_pert_names) & set(pert_onehot_map.keys())
    logger.info(f"Overlap between AnnData and data module: {len(overlap)} perturbations")
    if len(overlap) < len(unique_pert_names):
        missing = set(unique_pert_names) - set(pert_onehot_map.keys())
        logger.warning(f"Missing perturbations: {list(missing)[:10]}")

    # Check if there's a control perturbation that might match
    control_pert = cfg["data"]["kwargs"]["control_pert"]
    if args.pert_col == "drugname_drugconc":  # quick hack for tahoe
        control_pert = "[('DMSO_TF', 0.0, 'uM')]"
    logger.info(f"Control perturbation in data module: '{control_pert}'")

    matched_count = 0
    for idx, name in enumerate(pert_names):
        if name in pert_onehot_map:
            pert_tensor[idx] = pert_onehot_map[name]
            matched_count += 1
        else:
            # For now, use control perturbation as fallback
            if control_pert in pert_onehot_map:
                pert_tensor[idx] = pert_onehot_map[control_pert]
            else:
                # Use first available perturbation as fallback
                first_pert = list(pert_onehot_map.keys())[0]
                pert_tensor[idx] = pert_onehot_map[first_pert]

    logger.info(f"Matched {matched_count} out of {len(pert_names)} perturbations")

    # Process in batches with progress bar
    # Use cell_sentence_len as batch size since model expects this
    n_samples = len(pert_names)
    batch_size = cell_sentence_len  # Model requires this exact batch size
    n_batches = (n_samples + batch_size - 1) // batch_size  # Ceiling division

    logger.info(
        f"Running inference on {n_samples} samples in {n_batches} batches of size {batch_size} (model's cell_sentence_len)..."
    )

    all_preds = []

    with torch.no_grad():
        progress_bar = tqdm(total=n_samples, desc="Processing samples", unit="samples")

        for batch_idx in range(n_batches):
            start_idx = batch_idx * batch_size
            end_idx = min(start_idx + batch_size, n_samples)
            current_batch_size = end_idx - start_idx

            # Get batch data
            X_batch = torch.tensor(X[start_idx:end_idx], dtype=torch.float32).to(device)
            pert_batch = pert_tensor[start_idx:end_idx].to(device)
            pert_names_batch = pert_names[start_idx:end_idx].tolist()

            # Pad the batch to cell_sentence_len if it's the last incomplete batch
            if current_batch_size < cell_sentence_len:
                # Pad with zeros for embeddings
                padding_size = cell_sentence_len - current_batch_size
                X_pad = torch.zeros((padding_size, X_batch.shape[1]), device=device)
                X_batch = torch.cat([X_batch, X_pad], dim=0)

                # Pad perturbation tensor with control perturbation
                pert_pad = torch.zeros((padding_size, pert_batch.shape[1]), device=device)
                if control_pert in pert_onehot_map:
                    pert_pad[:] = pert_onehot_map[control_pert].to(device)
                else:
                    pert_pad[:, 0] = 1  # Default to first perturbation
                pert_batch = torch.cat([pert_batch, pert_pad], dim=0)

                # Extend perturbation names
                pert_names_batch.extend([control_pert] * padding_size)

            # Prepare batch - use same format as working code
            batch = {
                "ctrl_cell_emb": X_batch,
                "pert_emb": pert_batch,  # Keep as 2D tensor
                "pert_name": pert_names_batch,
                "batch": torch.zeros((1, cell_sentence_len), device=device),  # Use (1, cell_sentence_len)
            }

            # Run inference on batch using padded=False like in working code
            batch_preds = model.predict_step(batch, batch_idx=batch_idx, padded=False)

            # Extract predictions from the dictionary returned by predict_step
            # Use gene decoder output if available, otherwise use latent predictions
            if "pert_cell_counts_preds" in batch_preds and batch_preds["pert_cell_counts_preds"] is not None:
                # Use gene space predictions (from decoder)
                pred_tensor = batch_preds["pert_cell_counts_preds"]
            else:
                # Use latent space predictions
                pred_tensor = batch_preds["preds"]

            # Only keep predictions for the actual samples (not padding)
            actual_preds = pred_tensor[:current_batch_size]
            all_preds.append(actual_preds.cpu().numpy())

            # Update progress bar
            progress_bar.update(current_batch_size)

        progress_bar.close()

    # Concatenate all predictions
    preds_np = np.concatenate(all_preds, axis=0)

    # Save predictions to AnnData
    pred_key = "model_preds"
    adata.obsm[pred_key] = preds_np
    output_path = args.output or args.adata.replace(".h5ad", "_with_preds.h5ad")
    adata.write_h5ad(output_path)
    logger.info(f"Saved predictions to {output_path} (in adata.obsm['{pred_key}'])")


def main():
    parser = argparse.ArgumentParser(description="Run inference on AnnData with a trained model checkpoint.")
    add_arguments_infer(parser)
    args = parser.parse_args()
    run_tx_infer(args)


if __name__ == "__main__":
    main()



from typing import Dict, Optional

import torch
from geomloss import SamplesLoss

from .base import PerturbationModel
from .utils import build_mlp, get_activation_class, get_transformer_backbone


class OldNeuralOTPerturbationModel(PerturbationModel):
    """
    This model:
      1) Projects basal expression and perturbation encodings into a shared latent space.
      2) Uses an OT-based distributional loss (energy, sinkhorn, etc.) from geomloss.
      3) Enables cells to attend to one another, learning a set-to-set function rather than
      a sample-to-sample single-cell map.
    """

    def __init__(
        self,
        input_dim: int,
        hidden_dim: int,
        output_dim: int,
        pert_dim: int,
        predict_residual: bool = True,
        distributional_loss: str = "energy",
        transformer_backbone_key: str = "GPT2",
        transformer_backbone_kwargs: dict = None,
        output_space: str = "gene",
        gene_dim: Optional[int] = None,
        **kwargs,
    ):
        """
        Args:
            input_dim: dimension of the input expression (e.g. number of genes or embedding dimension).
            hidden_dim: not necessarily used, but required by PerturbationModel signature.
            output_dim: dimension of the output space (genes or latent).
            pert_dim: dimension of perturbation embedding.
            gpt: e.g. "TranslationTransformerSamplesModel".
            model_kwargs: dictionary passed to that model's constructor.
            loss: choice of distributional metric ("sinkhorn", "energy", etc.).
            **kwargs: anything else to pass up to PerturbationModel or not used.
        """
        # Call the parent PerturbationModel constructor
        super().__init__(
            input_dim=input_dim,
            hidden_dim=hidden_dim,
            gene_dim=gene_dim,
            output_dim=output_dim,
            pert_dim=pert_dim,
            output_space=output_space,
            **kwargs,
        )

        # Save or store relevant hyperparams
        self.predict_residual = predict_residual
        self.n_encoder_layers = kwargs.get("n_encoder_layers", 2)
        self.n_decoder_layers = kwargs.get("n_decoder_layers", 2)
        self.activation_class = get_activation_class(kwargs.get("activation", "gelu"))
        self.transformer_backbone_key = transformer_backbone_key
        self.transformer_backbone_kwargs = transformer_backbone_kwargs
        self.distributional_loss = distributional_loss
        self.cell_sentence_len = self.transformer_backbone_kwargs["n_positions"]
        self.gene_dim = gene_dim

        # Build the distributional loss from geomloss
        self.loss_fn = SamplesLoss(loss=self.distributional_loss)
        # self.loss_fn = LearnableAlignmentLoss()

        # Build the underlying neural OT network
        self._build_networks()

    def _build_networks(self):
        """
        Here we instantiate the actual GPT2-based model or any neuralOT translator
        via your old get_model(model_key, model_kwargs) approach.
        """
        self.pert_encoder = build_mlp(
            in_dim=self.pert_dim,
            out_dim=self.hidden_dim,
            hidden_dim=self.hidden_dim,
            n_layers=self.n_encoder_layers,
            dropout=self.dropout,
            activation=self.activation_class,
        )

        # Map the input embedding to the hidden space
        self.basal_encoder = build_mlp(
            in_dim=self.input_dim,
            out_dim=self.hidden_dim,
            hidden_dim=self.hidden_dim,
            n_layers=self.n_encoder_layers,
            dropout=self.dropout,
            activation=self.activation_class,
        )

        self.transformer_backbone, self.transformer_model_dim = get_transformer_backbone(
            self.transformer_backbone_key,
            self.transformer_backbone_kwargs,
        )

        self.project_out = build_mlp(
            in_dim=self.hidden_dim,
            out_dim=self.output_dim,
            hidden_dim=self.hidden_dim,
            n_layers=self.n_decoder_layers,
            dropout=self.dropout,
            activation=self.activation_class,
        )

        print(self)

    def encode_perturbation(self, pert: torch.Tensor) -> torch.Tensor:
        """If needed, define how we embed the raw perturbation input."""
        return self.pert_encoder(pert)

    def encode_basal_expression(self, expr: torch.Tensor) -> torch.Tensor:
        """Define how we embed basal state input, if needed."""
        return self.basal_encoder(expr)

    def perturb(self, pert: torch.Tensor, basal: torch.Tensor) -> torch.Tensor:
        """
        Return the latent perturbed state given the perturbation and basal state.
        """
        pert_embedding = self.encode_perturbation(pert).unsqueeze(1)  # shape: [batch_size, 1, hidden_dim]
        control_cells = self.encode_basal_expression(basal).unsqueeze(1)  # shape: [batch_size, 1, hidden_dim]
        cls_input = torch.zeros_like(pert_embedding)  # shape: [batch_size, 1, hidden_dim]
        seq_input = torch.cat([pert_embedding, control_cells, cls_input], dim=1)  # shape: [batch_size, 3, hidden_dim]

        # forward pass + extract CLS last hidden state
        prediction = self.transformer_backbone(inputs_embeds=seq_input).last_hidden_state[:, -1]

        # add to basal if predicting residual
        if self.predict_residual:
            # treat the actual prediction as a residual sum to basal
            return prediction + control_cells.squeeze(1)
        else:
            return prediction

    def forward(self, batch: dict) -> torch.Tensor:
        """
        The main forward call.
        """
        prediction = self.perturb(batch["pert_emb"], batch["ctrl_cell_emb"])
        output = self.project_out(prediction)

        return output

    def training_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> torch.Tensor:
        """Training step logic for both main model and decoder."""
        # Get model predictions (in latent space)
        pred = self(batch)
        pred = pred.reshape(-1, self.cell_sentence_len, self.output_dim)
        # TODO: please improve this, do not assume self.cell_sentence_len for this model
        target = batch["pert_cell_emb"]
        target = target.reshape(-1, self.cell_sentence_len, self.output_dim)
        main_loss = self.loss_fn(pred, target).mean()
        self.log("train_loss", main_loss)

        # Process decoder if available
        decoder_loss = None
        if self.gene_decoder is not None and "pert_cell_counts" in batch:
            # Train decoder to map latent predictions to gene space
            with torch.no_grad():
                latent_preds = pred.detach()  # Detach to prevent gradient flow back to main model

            pert_cell_counts_preds = self.gene_decoder(latent_preds)
            gene_targets = batch["pert_cell_counts"]
            gene_targets = gene_targets.reshape(-1, self.cell_sentence_len, self.gene_dim)
            decoder_loss = self.loss_fn(pert_cell_counts_preds, gene_targets).mean()

            # Log decoder loss
            self.log("decoder_loss", decoder_loss)

            total_loss = main_loss + decoder_loss
        else:
            total_loss = main_loss

        return total_loss

    def validation_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> None:
        """Validation step logic."""
        pred = self(batch)
        pred = pred.reshape(-1, self.cell_sentence_len, self.output_dim)
        target = batch["pert_cell_emb"]
        target = target.reshape(-1, self.cell_sentence_len, self.output_dim)
        loss = self.loss_fn(pred, target).mean()
        self.log("val_loss", loss)

        return {"loss": loss, "predictions": pred}

    def on_validation_batch_end(self, outputs, batch, batch_idx, dataloader_idx=0) -> None:
        """Track decoder performance during validation without training it."""
        if self.gene_decoder is not None and "pert_cell_counts" in batch:
            # Get model predictions from validation step
            latent_preds = outputs["predictions"]

            # Train decoder to map latent predictions to gene space
            pert_cell_counts_preds = self.gene_decoder(latent_preds)  # verify this is automatically detached
            gene_targets = batch["pert_cell_counts"]

            # Get decoder predictions
            pert_cell_counts_preds = pert_cell_counts_preds.reshape(-1, self.cell_sentence_len, self.gene_dim)
            gene_targets = gene_targets.reshape(-1, self.cell_sentence_len, self.gene_dim)
            decoder_loss = self.loss_fn(pert_cell_counts_preds, gene_targets).mean()

            # Log the validation metric
            self.log("decoder_val_loss", decoder_loss)

    def test_step(self, batch: Dict[str, torch.Tensor], batch_idx: int) -> None:
        pred = self.forward(batch, padded=False)
        target = batch["pert_cell_emb"]
        pred = pred.reshape(1, -1, self.output_dim)
        target = target.reshape(1, -1, self.output_dim)
        loss = self.loss_fn(pred, target).mean()
        self.log("test_loss", loss)
        pred = pred.reshape(-1, self.output_dim)
        target = target.reshape(-1, self.output_dim)

    def predict_step(self, batch, batch_idx, padded=True, **kwargs):
        """
        Typically used for final inference. We'll replicate old logic:
         returning 'preds', 'X', 'pert_name', etc.
        """
        latent_output = self.forward(batch)  # shape [B, ...]
        output_dict = {
            "preds": latent_output,
            "pert_cell_emb": batch.get("pert_cell_emb", None),
            "pert_cell_counts": batch.get("pert_cell_counts", None),
            "pert_name": batch.get("pert_name", None),
            "celltype_name": batch.get("cell_type", None),
            "batch": batch.get("batch", None),
            "ctrl_cell_emb": batch.get("ctrl_cell_emb", None),
        }

        if self.gene_decoder is not None:
            pert_cell_counts_preds = self.gene_decoder(latent_output)
            output_dict["pert_cell_counts_preds"] = pert_cell_counts_preds

        return output_dict



from typing import List, Literal, Optional, Union

import torch
from torchmetrics.functional import pearson_corrcoef, r2_score

from ..base import PerturbationModel
from .generation_model import ChemicalTransformerGenerator, TransformerGenerator
from .loss import masked_mse_loss
from .utils import map_raw_id_to_vocab_id


class scGPTForPerturbation(PerturbationModel):
    def __init__(
        self,
        ntoken: int,
        d_model: int,
        nhead: int,
        d_hid: int,
        nlayers: int,
        nlayers_cls: int,
        n_cls: int,
        n_drug_tokens: int = 0,
        dropout: float = 0.5,
        pad_token_id: int = 0,
        pad_value: int = 0,
        pert_pad_id: int = 2,
        do_mvc: bool = False,
        domain_spec_batchnorm: Union[bool, str] = False,
        cell_emb_style: str = "cls",
        mvc_decoder_style: str = "inner product",
        ecs_threshold: float = 0.3,
        explicit_zero_prob: bool = False,
        use_fast_transformer: bool = False,
        fast_transformer_backend: str = "flash",
        pre_norm: bool = False,
        lr: float = 1e-4,
        step_size_lr: int = 1,
        include_zero_gene: Optional[Literal["all", "batch-wise"]] = "all",
        max_seq_len: int = 1536,
        do_CLS: bool = True,
        do_CCE: bool = False,
        do_MVC: bool = False,
        do_ECS: bool = False,
        gene_names: Optional[List[str]] = None,
        embed_key: Optional[str] = None,
        perturbation_type: Literal["chemical", "genetic"] = "chemical",
        **kwargs,
    ):
        super().__init__(
            input_dim=None,
            hidden_dim=None,
            output_dim=None,
            pert_dim=None,
            dropout=dropout,
            lr=lr,
            loss_fn="mse",
            embed_key=embed_key,
            output_space="gene",
            gene_names=gene_names,
            batch_size=64,
        )
        self.ntoken = ntoken
        self.d_model = d_model
        self.nhead = nhead
        self.d_hid = d_hid
        self.nlayers = nlayers
        self.nlayers_cls = nlayers_cls
        self.n_cls = n_cls
        self.n_drug_tokens = n_drug_tokens
        self.dropout = dropout
        self.pad_token_id = pad_token_id
        self.pad_value = pad_value
        self.pert_pad_id = pert_pad_id
        self.do_mvc = do_mvc
        self.domain_spec_batchnorm = domain_spec_batchnorm
        self.cell_emb_style = cell_emb_style
        self.mvc_decoder_style = mvc_decoder_style
        self.ecs_threshold = ecs_threshold
        self.explicit_zero_prob = explicit_zero_prob
        self.use_fast_transformer = use_fast_transformer
        self.fast_transformer_backend = fast_transformer_backend
        self.pre_norm = pre_norm
        self.perturbation_type = perturbation_type.lower()

        for k, v in kwargs.items():
            print(f"WARNING: scGPTForPerturbation Model unused kwarg: {k}")

        self.lr = lr
        self.step_size_lr = step_size_lr
        self.include_zero_gene = include_zero_gene
        self.max_seq_len = max_seq_len
        self.do_CLS = do_CLS
        self.do_CCE = do_CCE
        self.do_MVC = do_MVC
        self.do_ECS = do_ECS

        assert self.perturbation_type in ["chemical", "genetic"], (
            "perturbation_type must be either 'chemical' or 'genetic'"
        )

        if self.perturbation_type == "chemical":
            assert self.n_drug_tokens > 0, "n_drug_tokens must be greater than 0 for chemical perturbation"

        self.save_hyperparameters()

        self.validation_outputs = []

        self._build_networks()

    def _build_networks(self):
        generator_params = dict(
            ntoken=self.ntoken,
            d_model=self.d_model,
            nhead=self.nhead,
            d_hid=self.d_hid,
            nlayers=self.nlayers,
            nlayers_cls=self.nlayers_cls,
            n_cls=self.n_cls,
            dropout=self.dropout,
            pad_token_id=self.pad_token_id,
            pad_value=self.pad_value,
            pert_pad_id=self.pert_pad_id,
            do_mvc=self.do_mvc,
            domain_spec_batchnorm=self.domain_spec_batchnorm,
            cell_emb_style=self.cell_emb_style,
            mvc_decoder_style=self.mvc_decoder_style,
            ecs_threshold=self.ecs_threshold,
            explicit_zero_prob=self.explicit_zero_prob,
            use_fast_transformer=self.use_fast_transformer,
            fast_transformer_backend=self.fast_transformer_backend,
            pre_norm=self.pre_norm,
        )
        if self.perturbation_type == "chemical":
            self.model = ChemicalTransformerGenerator(
                n_drug_tokens=self.n_drug_tokens,
                **generator_params,
            )
        elif self.perturbation_type == "genetic":
            self.model = TransformerGenerator(**generator_params)

    def encode_perturbation(self, pert: torch.Tensor) -> torch.Tensor:
        """Map perturbation to an effect vector in embedding space."""
        raise NotImplementedError("Perturbation encoding not supported for scGPT model")

    def encode_basal_expression(self, expr: torch.Tensor) -> torch.Tensor:
        """Expression is already in embedding space, pass through."""
        raise NotImplementedError("Basal expression encoding not supported for scGPT model")

    def perturb(self, pert: torch.Tensor, basal: torch.Tensor) -> torch.Tensor:
        """
        Given a perturbation and basal embeddings, compute the perturbed embedding.
        """
        # Project perturbation and basal cell state to latent space
        raise NotImplementedError("Perturb function not supported for scGPT model")

    def configure_optimizers(self):
        optimizer = torch.optim.Adam(self.parameters(), lr=self.lr)
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, self.step_size_lr, gamma=0.9)
        return [optimizer], [scheduler]

    def shared_step(
        self,
        batch,
        truncate=True,
    ):
        x_basal = batch["ctrl_cell_emb"]  # (batch_size, n_genes)
        x_pert = batch["pert_cell_emb"]  # (batch_size, n_genes)
        pert_ids = batch["pert_emb"].argmax(dim=1)
        if self.perturbation_type == "chemical":
            pert_flags = torch.zeros_like(x_pert, dtype=torch.long)  # no genes are perturbed
        else:
            pert_flags = batch["pert_flags"]  # (batch_size, n_genes)

        gene_ids = batch["gene_ids"][0]  # (n_genes,)

        nonpad_genes_mask = gene_ids != self.pad_token_id

        x_basal = x_basal[:, nonpad_genes_mask]
        x_pert = x_pert[:, nonpad_genes_mask]
        gene_ids = gene_ids[nonpad_genes_mask]
        pert_flags = pert_flags[:, nonpad_genes_mask]

        batch_size, n_genes = x_basal.size()

        if self.include_zero_gene == "all":
            input_gene_ids = torch.arange(n_genes, device=x_basal.device, dtype=torch.long)
        else:
            input_gene_ids = x_basal.nonzero()[:, 1].flatten().unique().sort()[0]  # TODO: need double-check

        # sample input_gene_id
        if truncate:
            if len(input_gene_ids) > self.max_seq_len:
                input_gene_ids = torch.randperm(len(input_gene_ids), device=x_basal.device)[: self.max_seq_len]

            x_basal = x_basal[:, input_gene_ids]
            x_pert = x_pert[:, input_gene_ids]
            pert_flags = pert_flags[:, input_gene_ids]

        mapped_input_gene_ids = map_raw_id_to_vocab_id(input_gene_ids, gene_ids)
        mapped_input_gene_ids = mapped_input_gene_ids.repeat(batch_size, 1)  # (batch_size, max_seq_len)

        src_key_padding_mask = mapped_input_gene_ids.eq(self.pad_token_id)
        # src_key_padding_mask = torch.zeros_like(
        #     x_basal, dtype=torch.bool, device=x_basal.device
        # )

        if self.perturbation_type == "genetic":
            pert_ids = None

        output_dict = self.model(
            mapped_input_gene_ids.long(),
            pert_ids,
            x_basal,
            pert_flags.long(),
            src_key_padding_mask=src_key_padding_mask,
            CLS=self.do_CLS and self.training,
            CCE=self.do_CCE and self.training,
            MVC=self.do_MVC and self.training,
            ECS=self.do_ECS and self.training,
        )
        output_values = output_dict["mlm_output"]  # batch_size, max_seq_len

        masked_positions = torch.ones_like(x_basal, dtype=torch.bool)  # Use all genes
        loss = masked_mse_loss(output_values, x_pert, masked_positions)

        output_dict["x_pred"] = output_dict["mlm_output"].float()
        output_dict["x_true"] = x_pert
        output_dict["x_basal"] = x_basal

        return loss, output_dict

    def compute_metrics(self, x_basal, x_pred, x_true):
        """
        Computes a sets of evaluation metrics for assessing perturbation response prediction's quality.

        Parameters
        ----------
        x_pred : torch.Tensor of shape (batch_size, n_genes)
            The predicted perturbation response.
        x_true : torch.Tensor of shape (batch_size, n_genes)
            The true perturbation response.
        """
        r2_mean = r2_score(x_pred.mean(0), x_true.mean(0))  # R2 score for mean gene expression

        change_pred = x_pred - x_basal
        change_true = x_true - x_basal

        pearson_mean_lfc = pearson_corrcoef(change_pred.mean(0), change_true.mean(0))

        return {
            "r2_mean": r2_mean,
            "pearson_mean_lfc": pearson_mean_lfc,
        }

    def training_step(self, batch, batch_idx):
        loss, batch_outputs = self.shared_step(batch)

        metrics = self.compute_metrics(
            x_basal=batch_outputs["x_basal"],
            x_pred=batch_outputs["x_pred"],
            x_true=batch_outputs["x_true"],
        )

        self.log("loss", loss, prog_bar=True)

        for k, v in metrics.items():
            self.log(k, v, prog_bar=True)

        return loss

    def validation_step(self, batch, batch_idx):
        loss, batch_outputs = self.shared_step(batch)

        metrics = self.compute_metrics(
            x_basal=batch_outputs["x_basal"],
            x_pred=batch_outputs["x_pred"],
            x_true=batch_outputs["x_true"],
        )

        # self.validation_outputs.append(
        #     (
        #         batch_outputs["x_basal"].detach().cpu(),
        #         batch_outputs["x_pred"].detach().cpu(),
        #         batch_outputs["x_true"].detach().cpu(),
        #     )
        # )

        self.log("val_loss", loss, prog_bar=True)

        for k, v in metrics.items():
            self.log(f"val_{k}", v, prog_bar=True)

        return loss

    # def validation_epoch_end(self, outputs):
    #     pass

    @torch.no_grad()
    def predict(self, batch):
        loss, batch_outputs = self.shared_step(batch, truncate=False)

        x_pred = batch_outputs["x_pred"]
        return x_pred

    def predict_step(self, batch, batch_idx, **kwargs):
        """
        Typically used for final inference. We'll replicate old logic:
         returning 'pred', 'X', 'pert_name', etc.
        """
        loss, batch_outputs = self.shared_step(batch, truncate=False)

        return {
            "preds": batch_outputs["x_pred"].float(),
            "pert_cell_emb": batch_outputs["x_true"].float(),
            "X_gene": batch_outputs["x_true"].float(),
            "pert_emb": batch.get("pert_emb", None),
            "pert_name": batch.get("pert_name", None),
            "cell_type": batch.get("cell_type", None),
            "batch": batch.get("batch_name", None),
            "ctrl_cell_emb": batch_outputs["x_basal"].float(),
        }



from typing import Union

import numpy as np
import torch


def map_raw_id_to_vocab_id(
    raw_ids: Union[np.ndarray, torch.Tensor],
    gene_ids: np.ndarray,
) -> Union[np.ndarray, torch.Tensor]:
    """
    Map some raw ids which are indices of the raw gene names to the indices of the

    Args:
        raw_ids: the raw ids to map
        gene_ids: the gene ids to map to
    """
    if isinstance(raw_ids, torch.Tensor):
        device = raw_ids.device
        dtype = raw_ids.dtype
        return_pt = True
        raw_ids = raw_ids.cpu().numpy()

    elif isinstance(raw_ids, np.ndarray):
        return_pt = False
        dtype = raw_ids.dtype

    else:
        raise ValueError("raw_ids must be either torch.Tensor or np.ndarray.")

    if raw_ids.ndim != 1:
        raise ValueError(f"raw_ids must be 1d, got {raw_ids.ndim}d.")

    if gene_ids.ndim != 1:
        raise ValueError(f"gene_ids must be 1d, got {gene_ids.ndim}d.")

    mapped_ids: np.ndarray = gene_ids[raw_ids]
    assert mapped_ids.shape == raw_ids.shape
    if return_pt:
        if isinstance(mapped_ids, np.ndarray):
            return torch.from_numpy(mapped_ids).type(dtype).to(device)
        return mapped_ids.to(dtype)
    return mapped_ids.astype(dtype)



import argparse as ap

from omegaconf import DictConfig, OmegaConf


def add_arguments_train(parser: ap.ArgumentParser):
    # Allow remaining args to be passed through to Hydra
    parser.add_argument("hydra_overrides", nargs="*", help="Hydra configuration overrides (e.g., data.batch_size=32)")


def run_tx_train(cfg: DictConfig):
    import json
    import logging
    import os
    import pickle
    import shutil
    from os.path import exists, join
    from pathlib import Path

    import lightning.pytorch as pl
    import torch
    from cell_load.data_modules import PerturbationDataModule
    from cell_load.utils.modules import get_datamodule
    from lightning.pytorch.loggers import WandbLogger
    from lightning.pytorch.plugins.precision import MixedPrecision

    from ...sets.callbacks import BatchSpeedMonitorCallback
    from ...sets.utils import get_checkpoint_callbacks, get_lightning_module, get_loggers

    logger = logging.getLogger(__name__)
    torch.set_float32_matmul_precision("medium")

    cfg_yaml = OmegaConf.to_yaml(cfg, resolve=True)
    cfg = OmegaConf.to_container(cfg, resolve=True)

    # Setup output directory
    run_output_dir = join(cfg["output_dir"], cfg["name"])
    if os.path.exists(run_output_dir) and cfg["overwrite"]:
        print(f"Output dir {run_output_dir} already exists, overwriting")
        shutil.rmtree(run_output_dir)
    os.makedirs(run_output_dir, exist_ok=True)

    # Set up wandb directory if needed
    if cfg["use_wandb"]:
        os.makedirs(cfg["wandb"]["local_wandb_dir"], exist_ok=True)

    with open(join(run_output_dir, "config.yaml"), "w") as f:
        f.write(cfg_yaml)

    # Set random seeds
    pl.seed_everything(cfg["training"]["train_seed"])

    # if the provided pert_col is drugname_drugconc, hard code the value of control pert
    # this is because it's surprisingly hard to specify a list of tuples in the config as a string
    if cfg["data"]["kwargs"]["pert_col"] == "drugname_drugconc":
        cfg["data"]["kwargs"]["control_pert"] = "[('DMSO_TF', 0.0, 'uM')]"

    # Initialize data module. this is backwards compatible with previous configs
    try:
        sentence_len = cfg["model"]["cell_set_len"]
    except KeyError:
        if cfg["model"]["name"].lower() in ["cpa", "scvi"] or cfg["model"]["name"].lower().startswith("scgpt"):
            if "cell_sentence_len" in cfg["model"]["kwargs"] and cfg["model"]["kwargs"]["cell_sentence_len"] > 1:
                sentence_len = cfg["model"]["kwargs"]["cell_sentence_len"]
                cfg["training"]["batch_size"] = 1
            else:
                sentence_len = 1
        else:
            try:
                sentence_len = cfg["model"]["kwargs"]["transformer_backbone_kwargs"]["n_positions"]
            except:
                sentence_len = cfg["model"]["kwargs"]["transformer_backbone_kwargs"]["max_position_embeddings"]

    if cfg["model"]["name"].lower().startswith("scgpt"):  # scGPT uses log-normalized expression
        cfg["data"]["kwargs"]["transform"] = "log-normalize"
        cfg["data"]["kwargs"]["hvg_names_uns_key"] = (
            "hvg_names" if cfg["data"]["kwargs"]["train_task"] != "replogle" else None
        )  # TODO: better to not hardcode this

        cfg["data"]["kwargs"]["dataset_cls"] = "scGPTPerturbationDataset"

        model_dir = Path(cfg["model"]["kwargs"]["pretrained_path"])

        vocab_file = model_dir / "vocab.json"

        vocab = json.load(open(vocab_file, "r"))
        cfg["model"]["kwargs"]["pad_token_id"] = vocab["<pad>"]
        for s in cfg["model"]["kwargs"]["special_tokens"]:
            if s not in vocab:
                vocab[s] = len(vocab)

        cfg["data"]["kwargs"]["vocab"] = vocab
        cfg["data"]["kwargs"]["perturbation_type"] = cfg["model"]["kwargs"]["perturbation_type"]
        cfg["model"]["kwargs"]["ntoken"] = len(vocab)
        cfg["model"]["kwargs"]["d_model"] = cfg["model"]["kwargs"]["embsize"]

        logger.info("Added vocab and hvg_names_uns_key to data kwargs for scGPT")

    elif cfg["model"]["name"].lower() == "cpa" and cfg["model"]["kwargs"]["recon_loss"] == "gauss":
        cfg["data"]["kwargs"]["transform"] = "log-normalize"
    elif cfg["model"]["name"].lower() == "scvi":
        cfg["data"]["kwargs"]["transform"] = None

    data_module: PerturbationDataModule = get_datamodule(
        cfg["data"]["name"],
        cfg["data"]["kwargs"],
        batch_size=cfg["training"]["batch_size"],
        cell_sentence_len=sentence_len,
    )

    with open(join(run_output_dir, "data_module.torch"), "wb") as f:
        # TODO-Abhi: only save necessary data
        data_module.save_state(f)

    data_module.setup(stage="fit")
    dl = data_module.train_dataloader()
    print("num_workers:", dl.num_workers)
    print("batch size:", dl.batch_size)

    var_dims = data_module.get_var_dims()  # {"gene_dim": …, "hvg_dim": …}
    if cfg["data"]["kwargs"]["output_space"] == "gene":
        gene_dim = var_dims.get("hvg_dim", 2000)  # fallback if key missing
    else:
        gene_dim = var_dims.get("gene_dim", 2000)  # fallback if key missing
    latent_dim = var_dims["output_dim"]  # same as model.output_dim
    hidden_dims = cfg["model"]["kwargs"].get("decoder_hidden_dims", [1024, 1024, 512])

    decoder_cfg = dict(
        latent_dim=latent_dim,
        gene_dim=gene_dim,
        hidden_dims=hidden_dims,
        dropout=cfg["model"]["kwargs"].get("decoder_dropout", 0.1),
        residual_decoder=cfg["model"]["kwargs"].get("residual_decoder", False),
    )

    # tuck it into the kwargs that will reach the LightningModule
    cfg["model"]["kwargs"]["decoder_cfg"] = decoder_cfg

    # Save the onehot maps as pickle files instead of storing in config
    cell_type_onehot_map_path = join(run_output_dir, "cell_type_onehot_map.pkl")
    pert_onehot_map_path = join(run_output_dir, "pert_onehot_map.pt")
    batch_onehot_map_path = join(run_output_dir, "batch_onehot_map.pkl")
    var_dims_path = join(run_output_dir, "var_dims.pkl")

    with open(cell_type_onehot_map_path, "wb") as f:
        pickle.dump(data_module.cell_type_onehot_map, f)
    torch.save(data_module.pert_onehot_map, pert_onehot_map_path)
    with open(batch_onehot_map_path, "wb") as f:
        pickle.dump(data_module.batch_onehot_map, f)
    with open(var_dims_path, "wb") as f:
        pickle.dump(var_dims, f)

    if cfg["model"]["name"].lower() in ["cpa", "scvi"] or cfg["model"]["name"].lower().startswith("scgpt"):
        cfg["model"]["kwargs"]["n_cell_types"] = len(data_module.celltype_onehot_map)
        cfg["model"]["kwargs"]["n_perts"] = len(data_module.pert_onehot_map)
        cfg["model"]["kwargs"]["n_batches"] = len(data_module.batch_onehot_map)

    # Create model
    model = get_lightning_module(
        cfg["model"]["name"],
        cfg["data"]["kwargs"],
        cfg["model"]["kwargs"],
        cfg["training"],
        data_module.get_var_dims(),
    )

    print(
        f"Model created. Estimated params size: {sum(p.numel() * p.element_size() for p in model.parameters()) / 1024**3:.2f} GB"
    )
    loggers = get_loggers(
        output_dir=cfg["output_dir"],
        name=cfg["name"],
        wandb_project=cfg["wandb"]["project"],
        wandb_entity=cfg["wandb"]["entity"],
        local_wandb_dir=cfg["wandb"]["local_wandb_dir"],
        use_wandb=cfg["use_wandb"],
        cfg=cfg,
    )

    # If using wandb, store the run path in a text file for eval
    # that matches the old train_lightning.py logic
    for lg in loggers:
        if isinstance(lg, WandbLogger):
            wandb_info_path = os.path.join(run_output_dir, "wandb_path.txt")
            with open(wandb_info_path, "w") as f:
                f.write(lg.experiment.path)
            break

    # Set up callbacks
    ckpt_callbacks = get_checkpoint_callbacks(
        cfg["output_dir"],
        cfg["name"],
        cfg["training"]["val_freq"],
        cfg["training"].get("ckpt_every_n_steps", 4000),
    )
    # Add BatchSpeedMonitorCallback to log batches per second to wandb
    batch_speed_monitor = BatchSpeedMonitorCallback()
    callbacks = ckpt_callbacks + [batch_speed_monitor]

    logger.info("Loggers and callbacks set up.")

    if cfg["model"]["name"].lower().startswith("scgpt"):
        plugins = [
            MixedPrecision(
                precision="bf16-mixed",
                device="cuda",
            )
        ]
    else:
        plugins = []

    if torch.cuda.is_available():
        accelerator = "gpu"
    else:
        accelerator = "cpu"

    # Decide on trainer params
    trainer_kwargs = dict(
        accelerator=accelerator,
        devices=1,
        max_steps=cfg["training"]["max_steps"],  # for normal models
        check_val_every_n_epoch=None,
        val_check_interval=cfg["training"]["val_freq"],
        logger=loggers,
        plugins=plugins,
        callbacks=callbacks,
        gradient_clip_val=cfg["training"]["gradient_clip_val"] if cfg["model"]["name"].lower() != "cpa" else None,
    )

    # If it's SimpleSum, override to do exactly 1 epoch, ignoring `max_steps`.
    if cfg["model"]["name"].lower() == "celltypemean" or cfg["model"]["name"].lower() == "globalsimplesum":
        trainer_kwargs["max_epochs"] = 1  # do exactly one epoch
        # delete max_steps to avoid conflicts
        del trainer_kwargs["max_steps"]

    # Build trainer
    print(f"Building trainer with kwargs: {trainer_kwargs}")
    trainer = pl.Trainer(**trainer_kwargs)
    print("Trainer built successfully")

    # Load checkpoint if exists
    checkpoint_path = join(ckpt_callbacks[0].dirpath, "last.ckpt")
    if not exists(checkpoint_path):
        checkpoint_path = None
    else:
        logging.info(f"!! Resuming training from {checkpoint_path} !!")

    print(f"Model device: {next(model.parameters()).device}")
    print(f"CUDA memory allocated: {torch.cuda.memory_allocated() / 1024**3:.2f} GB")
    print(f"CUDA memory reserved: {torch.cuda.memory_reserved() / 1024**3:.2f} GB")

    logger.info("Starting trainer fit.")

    # if a checkpoint does not exist, start with the provided checkpoint
    # this is mainly used for pretrain -> finetune workflows
    manual_init = cfg["model"]["kwargs"].get("init_from", None)
    if checkpoint_path is None and manual_init is not None:
        print(f"Loading manual checkpoint from {manual_init}")
        checkpoint_path = manual_init
        device = torch.device("cuda" if torch.cuda.is_available() else "cpu")
        checkpoint = torch.load(checkpoint_path, map_location=device)
        model_state = model.state_dict()
        checkpoint_state = checkpoint["state_dict"]

        pert_encoder_weight_key = "pert_encoder.0.weight"
        if pert_encoder_weight_key in checkpoint_state:
            checkpoint_pert_dim = checkpoint_state[pert_encoder_weight_key].shape[1]
            if checkpoint_pert_dim != model.pert_dim:
                print(
                    f"pert_encoder input dimension mismatch: model.pert_dim = {model.pert_dim} but checkpoint expects {checkpoint_pert_dim}. Overriding model's pert_dim and rebuilding pert_encoder."
                )
                # Rebuild the pert_encoder with the new pert input dimension
                from ...sets.models.utils import build_mlp

                model.pert_encoder = build_mlp(
                    in_dim=model.pert_dim,
                    out_dim=model.hidden_dim,
                    hidden_dim=model.hidden_dim,
                    n_layers=model.n_encoder_layers,
                    dropout=model.dropout,
                    activation=model.activation_class,
                )

        # Filter out mismatched size parameters
        filtered_state = {}
        for name, param in checkpoint_state.items():
            if name in model_state:
                if param.shape == model_state[name].shape:
                    filtered_state[name] = param
                else:
                    print(
                        f"Skipping parameter {name} due to shape mismatch: checkpoint={param.shape}, model={model_state[name].shape}"
                    )
            else:
                print(f"Skipping parameter {name} as it doesn't exist in the current model")

        # Load the filtered state dict
        model.load_state_dict(filtered_state, strict=False)
        print("About to call trainer.fit() with manual checkpoint...")

        # Train - for clarity we pass None
        trainer.fit(
            model,
            datamodule=data_module,
            ckpt_path=None,
        )
        print("trainer.fit() completed with manual checkpoint")
    else:
        print(f"About to call trainer.fit() with checkpoint_path={checkpoint_path}")
        # Train
        trainer.fit(
            model,
            datamodule=data_module,
            ckpt_path=checkpoint_path,
        )
        print("trainer.fit() completed")

    print("Training completed, saving final checkpoint...")

    # at this point if checkpoint_path does not exist, manually create one
    checkpoint_path = join(ckpt_callbacks[0].dirpath, "final.ckpt")
    if not exists(checkpoint_path):
        trainer.save_checkpoint(checkpoint_path)



import os
import pandas as pd
import numpy as np

from scipy.stats import pearsonr
from sklearn.metrics.pairwise import cosine_similarity

from pathlib import Path
import uuid

import logging
import requests
import time

# Constants for gene mapping
GENE_NAME_ENSEMPLE_MAP = {"GATD3A": "ENSMUSG00000053329", "GATD3B": "ENSG00000160221"}


def convert_gene_symbols_to_ensembl_rest(gene_symbols, species="human"):
    server = "https://grch37.rest.ensembl.org"

    # Map species to its scientific name
    species_map = {"human": "homo_sapiens", "mouse": "mus_musculus", "rat": "rattus_norvegicus"}

    species_name = species_map.get(species.lower(), species)
    gene_to_ensembl = {}

    for symbol in gene_symbols:
        # Construct the URL for the API request
        ext = f"/lookup/symbol/{species_name}/{symbol}?"

        # Make the request
        r = requests.get(server + ext, headers={"Content-Type": "application/json"})

        # Check if the request was successful
        if r.status_code != 200:
            print(f"Failed to retrieve data for {symbol}: {r.status_code}")
            continue

        # Parse the JSON response
        decoded = r.json()

        # Extract the Ensembl ID
        if "id" in decoded:
            gene_to_ensembl[symbol] = decoded["id"]

        # Sleep briefly to avoid overloading the server
        time.sleep(0.1)

    return gene_to_ensembl


def convert_symbols_to_ensembl(adata):
    import mygene

    gene_symbols = adata.var_names.tolist()

    mg = mygene.MyGeneInfo()
    results = mg.querymany(gene_symbols, scopes="symbol", fields="ensembl.gene", species="human")

    symbol_to_ensembl = {}
    for result in results:
        if "ensembl" in result and not result.get("notfound", False):
            if isinstance(result["ensembl"], list):
                symbol_to_ensembl[result["query"]] = result["ensembl"][0]["gene"]
            else:
                symbol_to_ensembl[result["query"]] = result["ensembl"]["gene"]

    for symbol in gene_symbols:
        if symbol_to_ensembl.get(symbol) is None:
            sym_results = convert_gene_symbols_to_ensembl_rest([symbol])
            if len(sym_results) > 0:
                symbol_to_ensembl[symbol] = sym_results[symbol]
                logging.info(f"Converted {symbol} to {symbol_to_ensembl[symbol]} using REST API")

    logging.info("Done...")
    for symbol in gene_symbols:
        if symbol_to_ensembl.get(symbol) is None:
            logging.info(f"{symbol} -> {symbol_to_ensembl.get(symbol, np.nan)}")
            if symbol in GENE_NAME_ENSEMPLE_MAP:
                symbol_to_ensembl[symbol] = GENE_NAME_ENSEMPLE_MAP[symbol]

    # Add the remaining or errored ones manually
    symbol_to_ensembl["PBK"] = "ENSG00000168078"
    return [symbol_to_ensembl.get(symbol, np.nan) for symbol in gene_symbols]


def is_valid_uuid(val):
    try:
        uuid.UUID(str(val))
        return True
    except ValueError:
        return False


def get_embedding_cfg(cfg):
    return cfg.embeddings[cfg.embeddings.current]


def get_dataset_cfg(cfg):
    return cfg.dataset[cfg.dataset.current]


def compute_pearson_delta(pred, true, ctrl, ctrl_true):
    """
    pred, true, ctrl, ctrl_true are numpy arrays of shape [n_cells, n_genes],
    or you can pass means if you prefer.

    We'll compute the correlation of (pred.mean - ctrl.mean) with (true.mean - ctrl_true.mean).
    """
    return pearsonr(pred.mean(axis=0) - ctrl.mean(axis=0), true.mean(axis=0) - ctrl_true.mean(axis=0))[0]


def compute_perturbation_ranking_score(adata_pred, adata_real, pert_col="gene", ctrl_pert="non-targeting"):
    """
    1) compute mean perturbation effect for each perturbation in pred and real
    2) measure how well the real perturbation matches the predicted one by rank
    returns the mean normalized rank of the true perturbation
    """
    # Step 1: compute mean effects
    mean_real_effect = _compute_mean_perturbation_effect(adata_real, pert_col, ctrl_pert)
    mean_pred_effect = _compute_mean_perturbation_effect(adata_pred, pert_col, ctrl_pert)
    all_perts = mean_real_effect.index.values

    ranks = []
    for pert in all_perts:
        real_effect = mean_real_effect.loc[pert].values.reshape(1, -1)
        pred_effects = mean_pred_effect.values

        # Compute pairwise cosine similarities
        similarities = cosine_similarity(real_effect, pred_effects).flatten()
        # where is the true row? (the same index in `all_perts`)
        true_pert_index = np.where(all_perts == pert)[0][0]

        # Sort by descending similarity
        sorted_indices = np.argsort(similarities)[::-1]
        # rank of the correct one:
        rank_of_true_pert = np.where(sorted_indices == true_pert_index)[0][0]
        ranks.append(rank_of_true_pert)

    mean_rank = np.mean(ranks) / len(all_perts)
    return mean_rank


def _compute_mean_perturbation_effect(adata, pert_col="gene", ctrl_pert="non-targeting"):
    """
    Helper to compute the mean effect (abs difference from control) for each perturbation.
    Actually we do the absolute difference from control row.
    """
    # shape: adata.X is (#cells, #genes)
    X = adata.X
    if not isinstance(X, np.ndarray):
        X = X.toarray()
    df = pd.DataFrame(X)
    df[pert_col] = adata.obs[pert_col].values
    mean_df = df.groupby(pert_col).mean(numeric_only=True)
    # difference from control
    return np.abs(mean_df - mean_df.loc[ctrl_pert])


def get_latest_checkpoint(cfg):
    run_name = "exp_{0}_layers_{1}_dmodel_{2}_samples_{3}_max_lr_{4}_op_dim_{5}".format(
        cfg.experiment.name,
        cfg.model.nlayers,
        cfg.model.emsize,
        cfg.dataset.pad_length,
        cfg.optimizer.max_lr,
        cfg.model.output_dim,
    )

    if cfg.experiment.checkpoint.path is None:
        return run_name, None
    chk_dir = os.path.join(cfg.experiment.checkpoint.path, cfg.experiment.name)
    chk = os.path.join(chk_dir, "last.ckpt")
    # chk = os.path.join(chk_dir, 'exp_rda_mmd_counts_1024_layers_4_dmodel_512_samples_1024_max_lr_0.00024_op_dim_512-epoch=1-step=581000.ckpt')
    if not os.path.exists(chk) or len(chk) == 0:
        chk = None

    return run_name, chk


def compute_gene_overlap_cross_pert(DE_pred, DE_true, control_pert="non-targeting", k=50):
    all_overlaps = {}
    for c_gene in DE_pred.index:
        if c_gene == control_pert:
            continue
        all_overlaps[c_gene] = len(set(DE_true.loc[c_gene].values).intersection(set(DE_pred.loc[c_gene].values))) / k

    return all_overlaps


def parse_chk_info(chk):
    chk_filename = Path(chk)
    epoch = chk_filename.stem.split("_")[-1].split("-")[1].split("=")[1]
    steps = chk_filename.stem.split("_")[-1].split("-")[2].split("=")[1]

    return int(epoch), int(steps)


def get_shapes_dict(dataset_path, filter_by_species=None):
    datasets_df = pd.read_csv(dataset_path)

    if filter_by_species is not None:
        datasets_df = datasets_df[datasets_df["species"] == filter_by_species]

    sorted_dataset_names = sorted(datasets_df["names"])
    datasets_df = datasets_df.drop_duplicates()  ## TODO: there should be no duplicates

    de_data_available = "groupid_for_de" in datasets_df.columns

    shapes_dict = {}
    dataset_path_map = {}
    dataset_group_map = {}  # Name of the obs column to be used for retrieing DE scrores

    shapes_dict["dev_immune_mouse"] = (443697, 4786)
    shapes_dict["dev_immune_human"] = (34009, 5566)
    shapes_dict["intestinal_tract_human"] = (69668, 5192)
    shapes_dict["gtex_human"] = (18511, 7109)
    shapes_dict["gut_endoderm_mouse"] = (113043, 6806)
    shapes_dict["luca"] = (249591, 7196)
    shapes_dict.update(
        {
            "madissoon_novel_lung": (190728, 8000),
            "flores_cerebellum_human": (20232, 8000),
            "osuch_gut_human": (272310, 8000),
            "msk_ovarian_human": (929690, 8000),
            "htan_vmuc_dis_epi_human": (65084, 8000),
            "htan_vmuc_val_epi_human": (57564, 8000),
            "htan_vmuc_non_epi_human": (9099, 8000),
            "hao_pbmc_3p_human": (161764, 8000),
            "hao_pbmc_5p_human": (49147, 8000),
            "gao_tumors_human": (36111, 8000),
            "swabrick_breast_human": (92427, 8000),
            "wu_cryo_tumors_human": (105662, 8000),
            "cell_line_het_human": (53513, 8000),
            "bi_allen_metastasis_human": (27787, 8000),
            "zheng68k_human": (68579, 8000),
            "zheng68k_12k_human": (68579, 12000),
            "mouse_embryo_ct": (153597, 12000),
            "regev_gtex_heart": (36574, 8000),
            "tabula_sapiens_heart": (11505, 8000),
            "10k_pbmcs": (11990, 12000),
            "epo_ido": (35834, 12000),
            "tabula_sapiens_kidney": (9641, 8000),
            "tabula_microcebus_kidney": (14592, 8000),
            "tabula_muris_kidney": (2781, 8000),
            "tabula_muris_senis_kidney": (19610, 8000),
            "immune_human": (33506, 8000),
        }
    )

    shapes_dict["zyl_sanes_glaucoma_pig"] = (5901, 6819)
    shapes_dict["parkinsons_macaF"] = (1062, 5103)

    for row in datasets_df.iterrows():
        ngenes = row[1].num_genes
        ncells = row[1].num_cells
        name = row[1].names
        dataset_path_map[name] = row[1].path
        if de_data_available:
            dataset_group_map[name] = datasets_df.set_index("names").loc[name]["groupid_for_de"]
        else:
            # This is for backward compatibility with old datasets CSV
            dataset_group_map[name] = "leiden"

        if not np.isnan(ngenes):
            shapes_dict[name] = (int(ncells), int(ngenes))
        else:
            shapes_dict[name] = (int(ncells), 8000)

    return datasets_df, sorted_dataset_names, shapes_dict, dataset_path_map, dataset_group_map



import argparse as ap

from ._fit import add_arguments_fit, run_emb_fit
from ._transform import add_arguments_transform, run_emb_transform

__all__ = ["run_emb_fit", "run_emb_transform", "add_arguments_emb"]


def add_arguments_emb(parser: ap.ArgumentParser):
    """"""
    subparsers = parser.add_subparsers(required=True, dest="subcommand")
    add_arguments_fit(subparsers.add_parser("fit"))
    add_arguments_transform(subparsers.add_parser("transform"))



import lightning.pytorch as pl
from lightning.pytorch.callbacks import Callback
from torch.optim import Optimizer

from ..models import PerturbationModel
from .batch_speed_monitor import BatchSpeedMonitorCallback

__all__ = ["PerturbationModel", "BatchSpeedMonitorCallback"]


class GradNormCallback(Callback):
    """
    Logs the gradient norm.
    """

    def on_before_optimizer_step(
        self, trainer: "pl.Trainer", pl_module: "pl.LightningModule", optimizer: Optimizer
    ) -> None:
        pl_module.log("train/gradient_norm", gradient_norm(pl_module))


def gradient_norm(model):
    total_norm = 0.0
    for p in model.parameters():
        if p.grad is not None:
            param_norm = p.grad.detach().data.norm(2)
            total_norm += param_norm.item() ** 2
    total_norm = total_norm ** (1.0 / 2)
    return total_norm



from .generation_model import TransformerGenerator
from .lightning_model import scGPTForPerturbation
from .loss import criterion_neg_log_bernoulli, masked_mse_loss, masked_relative_error
from .utils import map_raw_id_to_vocab_id

__all__ = [
    "scGPTForPerturbation",
    "TransformerGenerator",
    "masked_mse_loss",
    "criterion_neg_log_bernoulli",
    "masked_relative_error",
    "map_raw_id_to_vocab_id",
]



import math
from collections import defaultdict

import lightning as L
import numpy as np
import torch
from torch import nn
from torch.optim.lr_scheduler import StepLR
from torchmetrics.functional import accuracy

from ._base_modules import FocalLoss
from ._module import CPAModule


class CPATrainer(L.LightningModule):
    def __init__(
        self,
        module: CPAModule,
        lr=5e-4,
        wd=1e-6,
        n_steps_pretrain_ae: int = None,
        n_epochs_pretrain_ae: int = None,
        n_steps_kl_warmup: int = None,
        n_epochs_kl_warmup: int = None,
        n_steps_adv_warmup: int = None,
        n_epochs_adv_warmup: int = None,
        adv_steps: int = 3,
        reg_adv: float = 1.0,
        pen_adv: float = 1.0,
        adv_lr=1e-3,
        adv_wd=1e-6,
        step_size_lr: int = 45,
        do_clip_grad: bool = False,
        gradient_clip_value: float = 3.0,
        adv_loss: str = "cce",
        check_val_every_n_epoch: int = 5,
        **kwargs,
    ):
        """Training plan for the CPA model"""
        super().__init__()

        self.module = module

        self.lr = float(lr)
        self.n_steps_kl_warmup = n_steps_kl_warmup
        self.n_epochs_kl_warmup = n_epochs_kl_warmup

        self.automatic_optimization = False

        self.wd = float(wd)

        self.n_perts = module.n_perts

        self.n_steps_pretrain_ae = n_steps_pretrain_ae
        self.n_epochs_pretrain_ae = n_epochs_pretrain_ae

        self.n_steps_adv_warmup = n_steps_adv_warmup
        self.n_epochs_adv_warmup = n_epochs_adv_warmup
        self.adv_steps = adv_steps

        self.reg_adv = reg_adv
        self.pen_adv = pen_adv

        self.adv_lr = float(adv_lr)
        self.adv_wd = float(adv_wd)

        self.step_size_lr = step_size_lr

        self.do_clip_grad = do_clip_grad
        self.gradient_clip_value = gradient_clip_value
        self.check_val_every_n_epoch = check_val_every_n_epoch

        self.metrics = [
            "recon_loss",
            "KL",
            "disnt_basal",
            "disnt_after",
            "r2_mean",
            "r2_var",
            "adv_loss",
            "penalty_adv",
            "adv_perts",
            "acc_perts",
            "penalty_perts",
        ]

        self.epoch_history = defaultdict(list)

        # TODO: remove unused
        # device = torch.device("cuda" if torch.cuda.is_available() else "cpu")

        self.adv_loss = adv_loss.lower()
        self.gamma = kwargs.get("gamma", 2.0)
        if self.adv_loss == "focal":
            self.adv_loss_fn = FocalLoss(gamma=self.gamma, reduction="mean")
        else:
            self.adv_loss_fn = nn.CrossEntropyLoss()

    @property
    def kl_weight(self):
        return 0.0

    @property
    def adv_lambda(self):
        slope = self.reg_adv
        if self.n_steps_adv_warmup:
            global_step = self.global_step

            if self.n_steps_pretrain_ae:
                global_step -= self.n_steps_pretrain_ae

            if global_step <= self.n_steps_adv_warmup:
                proportion = global_step / self.n_steps_adv_warmup
                return slope * proportion
            else:
                return slope
        elif self.n_epochs_adv_warmup is not None:
            current_epoch = self.current_epoch

            if self.n_epochs_pretrain_ae:
                current_epoch -= self.n_epochs_pretrain_ae

            if current_epoch <= self.n_epochs_adv_warmup:
                proportion = current_epoch / self.n_epochs_adv_warmup
                return slope * proportion
            else:
                return slope
        else:
            return slope

    @property
    def do_start_adv_training(self):
        if self.n_steps_pretrain_ae:
            return self.global_step > self.n_steps_pretrain_ae
        elif self.n_epochs_pretrain_ae:
            return self.current_epoch > self.n_epochs_pretrain_ae
        else:
            return True

    def adversarial_loss(self, batch, z_basal, compute_penalty=True):
        """Computes adversarial classification losses and regularizations"""
        if compute_penalty:
            z_basal = z_basal.requires_grad_(True)

        adv_logits = self.module.forward_adv(z_basal)
        perts = batch["pert_emb"].argmax(1)
        pert_logits = adv_logits["pert_logits"]

        pert_adv_loss = self.adv_loss_fn(pert_logits, perts.long())
        pert_acc = accuracy(
            pert_logits.argmax(1),
            perts.long().view(
                -1,
            ),
            average="macro",
            task="multiclass",
            num_classes=self.n_perts,
        )

        cell_types = batch["cell_type_onehot"].argmax(1)
        cell_types_logits = adv_logits["cell_type_logits"]
        cell_types_adv_loss = self.adv_loss_fn(cell_types_logits, cell_types.long())
        cell_types_acc = accuracy(
            cell_types_logits.argmax(1),
            cell_types.long().view(
                -1,
            ),
            average="macro",
            task="multiclass",
            num_classes=self.n_perts,
        )

        batch_ids = batch["batch"].argmax(1)
        batch_ids_logits = adv_logits["batch_logits"]
        batch_ids_adv_loss = self.adv_loss_fn(batch_ids_logits, batch_ids.long())
        batch_ids_acc = accuracy(
            batch_ids_logits.argmax(1),
            batch_ids.long().view(
                -1,
            ),
            average="macro",
            task="multiclass",
            num_classes=self.n_perts,
        )

        adv_loss = pert_adv_loss + cell_types_adv_loss + batch_ids_adv_loss
        adv_acc = (pert_acc + cell_types_acc + batch_ids_acc) / 3.0

        if compute_penalty:
            # Penalty losses
            cell_type_penalty = (
                torch.autograd.grad(
                    cell_types_logits.sum(),
                    z_basal,
                    create_graph=True,
                    retain_graph=True,
                    only_inputs=True,
                )[0]
                .pow(2)
                .mean()
            )

            batch_penalty = (
                torch.autograd.grad(
                    batch_ids_logits.sum(),
                    z_basal,
                    create_graph=True,
                    retain_graph=True,
                    only_inputs=True,
                )[0]
                .pow(2)
                .mean()
            )

            pert_penalty = (
                torch.autograd.grad(
                    pert_logits.sum(),
                    z_basal,
                    create_graph=True,
                    retain_graph=True,
                    only_inputs=True,
                )[0]
                .pow(2)
                .mean()
            )

            total_penalty = cell_type_penalty + batch_penalty + pert_penalty

        else:
            total_penalty = torch.tensor(0.0, device=z_basal.device)

        return adv_loss, adv_acc, total_penalty

    def configure_optimizers(self):
        ae_params = (
            list(filter(lambda p: p.requires_grad, self.module.encoder.parameters()))
            + list(filter(lambda p: p.requires_grad, self.module.decoder.parameters()))
            + list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.pert_embeddings.parameters(),
                )
            )
            + list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.cell_type_embeddings.parameters(),
                )
            )
            + list(filter(lambda p: p.requires_grad, self.module.batch_embeddings.parameters()))
        )

        if self.module.recon_loss in ["zinb", "nb"]:
            ae_params += [self.module.px_r]

        optimizer_autoencoder = torch.optim.Adam(ae_params, lr=self.lr, weight_decay=self.wd)

        scheduler_autoencoder = StepLR(optimizer_autoencoder, step_size=self.step_size_lr, gamma=0.9)

        adv_params = (
            list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.perturbation_classifier.parameters(),
                )
            )
            + list(
                filter(
                    lambda p: p.requires_grad,
                    self.module.cell_type_classifier.parameters(),
                )
            )
            + list(filter(lambda p: p.requires_grad, self.module.batch_classifier.parameters()))
        )

        optimizer_adversaries = torch.optim.Adam(adv_params, lr=self.adv_lr, weight_decay=self.adv_wd)
        scheduler_adversaries = StepLR(optimizer_adversaries, step_size=self.step_size_lr, gamma=0.9)

        optimizers = [optimizer_autoencoder, optimizer_adversaries]
        schedulers = [scheduler_autoencoder, scheduler_adversaries]

        if self.step_size_lr is not None:
            return optimizers, schedulers
        else:
            return optimizers

    def training_step(self, batch, batch_idx):
        opt, opt_adv = self.optimizers()

        enc_outputs, dec_outputs = self.module.forward(batch)

        recon_loss, kl_loss = self.module.loss(
            batch=batch,
            encoder_outputs=enc_outputs,
            decoder_outputs=dec_outputs,
        )

        if self.do_start_adv_training:
            if self.adv_steps is None:
                opt.zero_grad()

                z_basal = enc_outputs["z_basal"]

                adv_loss, adv_acc, adv_penalty = self.adversarial_loss(
                    batch=batch,
                    z_basal=z_basal,
                    compute_penalty=False,
                )

                loss = recon_loss + self.kl_weight * kl_loss - self.adv_lambda * adv_loss

                self.manual_backward(loss)

                if self.do_clip_grad:
                    self.clip_gradients(
                        opt,
                        gradient_clip_val=self.gradient_clip_value,
                        gradient_clip_algorithm="norm",
                    )

                opt.step()

                opt_adv.zero_grad()

                adv_loss, adv_acc, adv_penalty = self.adversarial_loss(
                    batch=batch,
                    z_basal=z_basal.detach(),
                    compute_penalty=True,
                )

                adv_loss = adv_loss + self.pen_adv * adv_penalty

                self.manual_backward(adv_loss)

                if self.do_clip_grad:
                    self.clip_gradients(
                        opt_adv,
                        gradient_clip_val=self.gradient_clip_value,
                        gradient_clip_algorithm="norm",
                    )

                opt_adv.step()

            elif batch_idx % self.adv_steps == 0:
                opt_adv.zero_grad()

                z_basal = enc_outputs["z_basal"]

                adv_loss, adv_acc, adv_penalty = self.adversarial_loss(
                    batch=batch,
                    z_basal=z_basal.detach(),
                    compute_penalty=True,
                )

                adv_loss = adv_loss + self.pen_adv * adv_penalty

                self.manual_backward(adv_loss)

                if self.do_clip_grad:
                    self.clip_gradients(
                        opt_adv,
                        gradient_clip_val=self.gradient_clip_value,
                        gradient_clip_algorithm="norm",
                    )

                opt_adv.step()

            # Model update
            else:
                opt.zero_grad()

                z_basal = enc_outputs["z_basal"]

                adv_loss, adv_acc, adv_penalty = self.adversarial_loss(
                    batch=batch,
                    z_basal=z_basal,
                    compute_penalty=False,
                )

                loss = recon_loss + self.kl_weight * kl_loss - self.adv_lambda * adv_loss

                self.manual_backward(loss)

                if self.do_clip_grad:
                    self.clip_gradients(
                        opt,
                        gradient_clip_val=self.gradient_clip_value,
                        gradient_clip_algorithm="norm",
                    )

                opt.step()
        else:
            opt.zero_grad()

            z_basal = enc_outputs["z_basal"]

            loss = recon_loss + self.kl_weight * kl_loss

            self.manual_backward(loss)

            if self.do_clip_grad:
                self.clip_gradients(
                    opt,
                    gradient_clip_val=self.gradient_clip_value,
                    gradient_clip_algorithm="norm",
                )

            opt.step()

            opt_adv.zero_grad()

            adv_loss, adv_acc, adv_penalty = self.adversarial_loss(
                batch=batch,
                z_basal=z_basal.detach(),
                compute_penalty=True,
            )

            adv_loss = adv_loss + self.pen_adv * adv_penalty

            self.manual_backward(adv_loss)

            if self.do_clip_grad:
                self.clip_gradients(
                    opt_adv,
                    gradient_clip_val=self.gradient_clip_value,
                    gradient_clip_algorithm="norm",
                )

            opt_adv.step()

        r2_mean, r2_lfc = self.module.r2_metric(batch, enc_outputs, dec_outputs)

        disnt_basal, disnt_after = self.module.disentanglement(batch, enc_outputs, dec_outputs)

        results = {
            "recon_loss": recon_loss.item(),
            "KL": kl_loss.item(),
            "r2_mean": r2_mean,
            "r2_lfc": r2_lfc,
            "adv_loss": adv_loss.item(),
            "adv_acc": adv_acc.item(),
            "penalty_adv": adv_penalty.item(),
            "es_metric": r2_mean + np.e ** (disnt_after - disnt_basal),
            "disnt_basal": disnt_basal,
            "disnt_after": disnt_after,
        }

        self.log(
            "recon_loss",
            recon_loss.item(),
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "r2_mean",
            r2_mean,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "r2_lfc",
            r2_lfc,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "adv_loss",
            adv_loss.item(),
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "disnt_basal",
            disnt_basal,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "disnt_after",
            disnt_after,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        self.log(
            "adv_acc",
            adv_acc,
            on_epoch=True,
            on_step=False,
            prog_bar=True,
            logger=True,
        )

        if self.global_step % self.step_size_lr * 1000 == 0:
            sch, sch_adv = self.lr_schedulers()
            sch.step()
            sch_adv.step()

        return results

    def validation_step(self, batch, batch_idx):
        enc_outputs, dec_outputs = self.module.forward(batch)

        recon_loss, kl_loss = self.module.loss(
            batch=batch,
            encoder_outputs=enc_outputs,
            decoder_outputs=dec_outputs,
        )

        r2_mean, r2_lfc = self.module.r2_metric(batch, enc_outputs, dec_outputs)

        disnt_basal, disnt_after = self.module.disentanglement(batch, enc_outputs, dec_outputs)

        self.log("val_r2_mean", r2_mean, prog_bar=True)
        self.log("val_r2_lfc", r2_lfc, prog_bar=True)
        self.log("es_metric", r2_mean + math.e ** (disnt_after - disnt_basal), prog_bar=True)

    def test_step(self, batch, batch_idx):
        enc_outputs, dec_outputs = self.module.forward(batch)

        recon_loss, kl_loss = self.module.loss(
            batch=batch,
            encoder_outputs=enc_outputs,
            decoder_outputs=dec_outputs,
        )

        r2_mean, r2_lfc = self.module.r2_metric(batch, enc_outputs, dec_outputs)

        self.log("test_recon", recon_loss.item(), prog_bar=True)
        self.log("test_r2_mean", r2_mean, prog_bar=True)
        self.log("test_r2_lfc", r2_lfc, prog_bar=True)

        x_pred = self.module.get_expression(batch, n_samples=1)

        return x_pred.detach().cpu().numpy()
